#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of wrap prescription v1."""

from flexmock import flexmock
import pytest
import yaml

from thoth.adviser.context import Context
from thoth.adviser.state import State
from thoth.adviser.prescription.v1 import WrapPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_WRAP_SCHEMA
from thoth.python import PackageVersion
from thoth.python import Source

from .base import AdviserUnitPrescriptionTestCase


class TestWrapPrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to wrap prescription v1."""

    def test_run_stack_info(self, context: Context, state: State) -> None:
        """Check assigning stack info."""
        prescription_str = """
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
 state:
    resolved_dependencies:
      - name: werkzeug
        version: "<=1.0.0"
        index_url: 'https://pypi.org/simple'
run:
  stack_info:
    - type: WARNING
      message: Some message
      link: https://pypi.org/project/werkzeug
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)
        state.add_resolved_dependency(("werkzeug", "0.5.0", "https://pypi.org/simple"))
        self.check_run_stack_info(context, WrapPrescription, state=state)

    @pytest.mark.parametrize("log_level", ["INFO", "ERROR", "WARNING"])
    def test_run_log(self, caplog, context: Context, state: State, log_level: str) -> None:
        """Check logging messages."""
        prescription_str = f"""
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
 state:
    resolved_dependencies:
      - name: flask
run:
  log:
    message: Seen flask in one of the resolved stacks
    type: {log_level}
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)
        state.add_resolved_dependency(("flask", "0.12", "https://pypi.org/simple"))
        self.check_run_log(caplog, context, log_level, WrapPrescription, state=state)

    def test_run_no_match(self, context: Context, state: State) -> None:
        """Test running this pipeline unit without match."""
        prescription_str = """
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: flask
        version: "<=1.0.0,>=0.12"
        index_url: "https://pypi.org/simple"
      - name: connexion
        version: "==2.7.0"
        index_url: "https://pypi.org/simple"
run:
  stack_info:
    - type: ERROR
      message: This message will not be shown
      link: https://pypi.org/project/connexion
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)
        state.add_resolved_dependency(("flask", "0.12", "https://pypi.org/simple"))
        state.add_resolved_dependency(("connexion", "2.0.0", "https://pypi.org/simple"))

        assert not context.stack_info

        unit = WrapPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        assert not context.stack_info

    def test_run_justification(self, context: Context, state: State) -> None:
        """Test running this pipeline unit without match."""
        prescription_str = """
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: flask
        version: "<=1.0.0,>=0.12"
        index_url: "https://pypi.org/simple"
      - name: connexion
        version: "==2.7.0"
        index_url: "https://pypi.org/simple"
run:
  justification:
    - type: ERROR
      message: This message will not be shown
      link: https://pypi.org/project/connexion
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)
        state.add_resolved_dependency(("flask", "0.12", "https://pypi.org/simple"))
        state.add_resolved_dependency(("connexion", "2.7.0", "https://pypi.org/simple"))

        state.justification.clear()

        unit = WrapPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        assert state.justification == unit.run_prescription["justification"]

    def test_should_include(self) -> None:
        """Test including this pipeline unit."""
        prescription_str = """
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: flask
        version: "<=1.0.0,>=0.12"
        index_url: "https://pypi.org/simple"
run:
  justification:
    - type: ERROR
      message: This message will not be shown
      link: https://pypi.org/project/connexion
"""
        flexmock(WrapPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(WrapPrescription.should_include(builder_context)) == [
            {
                "package_name": "flask",
                "match": {
                    "state": {
                        "resolved_dependencies": [
                            {
                                "name": "flask",
                                "version": "<=1.0.0,>=0.12",
                                "index_url": "https://pypi.org/simple",
                            }
                        ],
                    },
                },
                "prescription": {"run": False},
                "run": {
                    "justification": [
                        {
                            "type": "ERROR",
                            "message": "This message will not be shown",
                            "link": "https://pypi.org/project/connexion",
                        }
                    ]
                },
            }
        ]

    def test_should_include_multi(self) -> None:
        """Test including this pipeline unit multiple times."""
        prescription_str = """
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
  - state:
      resolved_dependencies:
        - name: tensorflow-cpu
  - state:
      resolved_dependencies:
        - name: tensorflow
  - state:
      resolved_dependencies:
        - name: intel-tensorflow
run:
  justification:
    - type: ERROR
      message: This message will be shown
      link: https://pypi.org/project/tensorflow
"""
        flexmock(WrapPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(WrapPrescription.should_include(builder_context)) == [
            {
                "package_name": "tensorflow-cpu",
                "match": {
                    "state": {
                        "resolved_dependencies": [
                            {
                                "name": "tensorflow-cpu",
                            }
                        ],
                    },
                },
                "prescription": {"run": False},
                "run": {
                    "justification": [
                        {
                            "type": "ERROR",
                            "message": "This message will be shown",
                            "link": "https://pypi.org/project/tensorflow",
                        }
                    ]
                },
            },
            {
                "package_name": "tensorflow",
                "match": {
                    "state": {
                        "resolved_dependencies": [
                            {
                                "name": "tensorflow",
                            }
                        ],
                    },
                },
                "prescription": {"run": False},
                "run": {
                    "justification": [
                        {
                            "type": "ERROR",
                            "message": "This message will be shown",
                            "link": "https://pypi.org/project/tensorflow",
                        }
                    ]
                },
            },
            {
                "package_name": "intel-tensorflow",
                "match": {
                    "state": {
                        "resolved_dependencies": [
                            {
                                "name": "intel-tensorflow",
                            }
                        ],
                    },
                },
                "prescription": {"run": False},
                "run": {
                    "justification": [
                        {
                            "type": "ERROR",
                            "message": "This message will be shown",
                            "link": "https://pypi.org/project/tensorflow",
                        }
                    ]
                },
            },
        ]

    def test_should_include_no_package_name(self) -> None:
        """Test including this pipeline unit without any specific resolved package."""
        prescription_str = """
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
run:
  justification:
    - type: ERROR
      message: This message will not be shown
      link: https://pypi.org/project/connexion
"""
        flexmock(WrapPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(WrapPrescription.should_include(builder_context)) == [
            {
                "package_name": None,
                "match": {},
                "prescription": {"run": False},
                "run": {
                    "justification": [
                        {
                            "type": "ERROR",
                            "message": "This message will not be shown",
                            "link": "https://pypi.org/project/connexion",
                        }
                    ]
                },
            },
        ]

    def test_no_should_include(self) -> None:
        """Test not including this pipeline unit."""
        prescription_str = """
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: flask
        version: "<=1.0.0,>=0.12"
        index_url: "https://pypi.org/simple"
run:
  justification:
    - type: ERROR
      message: This message will not be shown
      link: https://pypi.org/project/connexion
"""
        flexmock(WrapPrescription).should_receive("_should_include_base").replace_with(lambda _: False).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(WrapPrescription.should_include(builder_context)) == []

    def test_advised_manifest_changes(self, state: State, context: Context) -> None:
        """Test advising changes in the manifest files."""
        prescription_str = """
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: intel-tensorflow
run:
  advised_manifest_changes:
    apiVersion: apps.openshift.io/v1
    kind: DeploymentConfig
    patch:
      op: add
      path: /spec/template/spec/containers/0/env/0
      value:
        name: OMP_NUM_THREADS
        value: "1"
  stack_info:
  - type: INFO
    message: Added advised changes
    link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)
        state.add_resolved_dependency(("intel-tensorflow", "2.2.0", "https://pypi.org/simple"))

        state.justification.clear()
        assert state.advised_manifest_changes == []

        unit = WrapPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None
            # Run one more time to verify stack info and advised manifest changes are added just once.
            assert unit.run(state) is None

        assert state.advised_manifest_changes == [
            {
                "apiVersion": "apps.openshift.io/v1",
                "kind": "DeploymentConfig",
                "patch": {
                    "op": "add",
                    "path": "/spec/template/spec/containers/0/env/0",
                    "value": {"name": "OMP_NUM_THREADS", "value": "1"},
                },
            }
        ]
        assert context.stack_info == [
            {"type": "INFO", "message": "Added advised changes", "link": "https://thoth-station.ninja"}
        ]

    @pytest.mark.parametrize("develop", [True, False])
    def test_run_develop(self, context: Context, state: State, develop: bool) -> None:
        """Test running this pipeline unit based on matching develop flag."""
        prescription_str = f"""
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: flask
        develop: {'true' if develop else 'false'}
run:
  justification:
    - type: INFO
      message: This message will be shown
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)

        package_version = PackageVersion(
            name="flask",
            version="==2.0.1",
            index=Source("https://pypi.org/simple"),
            develop=develop,
        )
        state.add_resolved_dependency(package_version.to_tuple())
        context.register_package_version(package_version)

        state.justification.clear()

        unit = WrapPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None
            # Run one more time to verify justification is added only once.
            assert unit.run(state) is None

        assert state.justification == unit.run_prescription["justification"]

    @pytest.mark.parametrize("develop", [True, False])
    def test_run_develop_no_match(self, context: Context, state: State, develop: bool) -> None:
        """Test running this pipeline unit based on NOT matching develop flag."""
        prescription_str = f"""
name: WrapUnit
type: wrap
should_include:
  times: 1
  adviser_pipeline: true
match:
  state:
    resolved_dependencies:
      - name: flask
        develop: {'true' if develop else 'false'}
run:
  justification:
    - type: INFO
      message: This message will be shown
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_WRAP_SCHEMA(prescription)
        WrapPrescription.set_prescription(prescription)

        package_version = PackageVersion(
            name="flask",
            version="==2.0.1",
            index=Source("https://pypi.org/simple"),
            develop=not develop,
        )
        state.add_resolved_dependency(package_version.to_tuple())
        context.register_package_version(package_version)

        state.justification.clear()

        unit = WrapPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state) is None

        assert not state.justification



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Handle caching of pipeline units."""

import functools
import logging
from typing import Any
from typing import Dict
from typing import Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable
    from .unit import UnitPrescription  # noqa: F401
    from ...pipeline_builder import PipelineBuilderContext

    SHOULD_INCLUDE_FUNC_TYPE = Callable[[Type["UnitPrescription"], str, "PipelineBuilderContext", Dict[str, Any]], bool]

_LOGGER = logging.getLogger(__name__)


def should_include_cache(func: "SHOULD_INCLUDE_FUNC_TYPE") -> "SHOULD_INCLUDE_FUNC_TYPE":  # noqa: D202
    """Handle caching for parts of prescription units that do not change during pipeline construction."""

    @functools.wraps(func)
    def wrapped(
        cls: Type["UnitPrescription"],
        unit_name: str,
        pipeline_builder: "PipelineBuilderContext",
        should_include_dict: Dict[str, Any],
    ) -> bool:
        cached_result = cls.SHOULD_INCLUDE_CACHE.get(unit_name)
        if cached_result is not None:
            _LOGGER.debug(
                "%s: Using pre-cached result (%r) of should include prescription part", unit_name, cached_result
            )
            return cached_result

        result: bool = func(cls, unit_name, pipeline_builder, should_include_dict)
        cls.SHOULD_INCLUDE_CACHE[unit_name] = result
        return result

    return wrapped



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Tests related to filtering out Python packages that are not part of a new release."""

import json

import pytest
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import PackageUpdateSieve
from thoth.adviser.context import Context
from thoth.common import get_justification_link as jl
from thoth.python import Source
from thoth.python import PackageVersion

from ..base import AdviserUnitTestCase


class TestPackageUpdateSieve(AdviserUnitTestCase):
    """Tests related to filtering out Python packages that are not part of a new release."""

    UNIT_TESTED = PackageUpdateSieve
    _PACKAGE_UPDATE_DICT = {
        "package_name": "tensorflow",
        "package_version": "2.5.0",
        "index_url": "https://pypi.org/simple",
    }

    @pytest.mark.skip(
        reason="Skip default configuration as default is not considered valid and needs further adjustments"
    )
    def test_default_configuration(self) -> None:
        """Skip testing default configuration."""

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.UNIT_TESTED._PACKAGE_UPDATE = json.dumps(self._PACKAGE_UPDATE_DICT)

        try:
            builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
            self.verify_multiple_should_include(builder_context)
        finally:
            self.UNIT_TESTED._PACKAGE_UPDATE = None

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test including this pipeline unit based on supplied package."""
        self.UNIT_TESTED._PACKAGE_UPDATE = json.dumps(self._PACKAGE_UPDATE_DICT)

        try:
            assert list(self.UNIT_TESTED.should_include(builder_context)) == [self._PACKAGE_UPDATE_DICT]
        finally:
            self.UNIT_TESTED._PACKAGE_UPDATE = None

    def test_should_include_no_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test NOT including this pipeline unit when no package is supplied."""
        assert self.UNIT_TESTED._PACKAGE_UPDATE is None

        try:
            assert list(self.UNIT_TESTED.should_include(builder_context)) == []
        finally:
            self.UNIT_TESTED._PACKAGE_UPDATE = None

    def test_sieve(self, context: Context) -> None:
        """Test removing packages that are not part of an update."""
        pypi = Source("https://pypi.org/simple")
        pv0 = PackageVersion(name="tensorflow", version="==2.5.0", index=pypi, develop=False)
        pv1 = PackageVersion(name="tensorflow", version="==2.4.0", index=pypi, develop=False)
        pv2 = PackageVersion(name="tensorflow", version="==2.3.0", index=pypi, develop=False)

        assert not context.stack_info

        self.UNIT_TESTED._PACKAGE_UPDATE = json.dumps(self._PACKAGE_UPDATE_DICT)
        try:
            unit = self.UNIT_TESTED()
            unit.update_configuration(self._PACKAGE_UPDATE_DICT)

            unit.pre_run()
            with self.UNIT_TESTED.assigned_context(context):
                assert list(unit.run(pv for pv in (pv0, pv1, pv2))) == [pv0]
        finally:
            self.UNIT_TESTED._PACKAGE_UPDATE = None

        assert context.stack_info == [
            {
                "link": jl("update"),
                "message": "Removing package ('tensorflow', '2.4.0', "
                "'https://pypi.org/simple') as advising on a new release",
                "package_name": "tensorflow",
                "type": "WARNING",
            },
            {
                "link": jl("update"),
                "message": "Removing package ('tensorflow', '2.3.0', "
                "'https://pypi.org/simple') as advising on a new release",
                "package_name": "tensorflow",
                "type": "WARNING",
            },
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of add package step."""

import yaml

import pytest

from thoth.adviser.context import Context
from thoth.adviser.state import State
from thoth.adviser.prescription.v1 import AddPackageStepPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages.exceptions import NotFoundError

from .base import AdviserUnitPrescriptionTestCase


class TestAddPackageStepPrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to add package step prescription."""

    @pytest.mark.parametrize(
        "package_version_from_version,package_version_from_index,package_version_from_develop,"
        "resolved_version,resolved_index,resolved_develop,"
        "add_resolved,pipeline_run",
        [
            ("<=0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "false", True, True),
            # Version not matching.
            (">0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "false", True, False),
            # Index not matching.
            (
                "<=0.25.0",
                '{"not": "https://pypi.org/simple"}',
                "false",
                "==8.0.0",
                "https://pypi.org/simple",
                "false",
                True,
                False,
            ),
            # Develop flag not matching.
            ("<=0.25.0", "https://pypi.org/simple", "true", "==8.0.0", "https://pypi.org/simple", "false", True, False),
            # Version not matching.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "!=8.0.0",
                "https://pypi.org/simple",
                "false",
                True,
                False,
            ),
            # Index not matching.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "==8.0.0",
                '{"not": "https://pypi.org/simple"}',
                "false",
                True,
                False,
            ),
            # Develop flag not matching.
            ("<=0.25.0", "https://pypi.org/simple", "false", "==8.0.0", "https://pypi.org/simple", "true", True, False),
            # Resolved dependency nit present.
            (
                "<=0.25.0",
                "https://pypi.org/simple",
                "false",
                "==8.0.0",
                "https://pypi.org/simple",
                "false",
                False,
                False,
            ),
        ],
    )
    def test_run_package_version_from_with_resolved(
        self,
        context: Context,
        state: State,
        package_version_from_version: str,
        package_version_from_index: str,
        package_version_from_develop: str,
        resolved_version: str,
        resolved_index: str,
        resolved_develop: str,
        add_resolved: bool,
        pipeline_run: bool,
    ) -> None:
        """Test running the add package unit."""
        prescription_str = f"""
name: AddPackageStep
type: step.AddPackage
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
  state:
    package_version_from:
    - name: scikit-learn
      version: "{package_version_from_version}"
      develop: {package_version_from_develop}
      index_url: {package_version_from_index}
    resolved_dependencies:
    - name: click
      version: "{resolved_version}"
      develop: {resolved_develop}
      index_url: {resolved_index}
run:
  package_version:
    name: daiquiri
    locked_version: ==2.0.0
    index_url: https://pypi.org/simple
    develop: true
  stack_info:
    - type: INFO
      message: "Hello, Thoth!"
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA(prescription)
        AddPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        package_version_resolved = PackageVersion(
            name="click",
            version="==8.0.0",
            index=pypi,
            develop=False,
        )

        assert "dauiqiri" not in state.resolved_dependencies

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        if add_resolved:
            state.add_resolved_dependency(package_version_resolved.to_tuple())
            context.register_package_version(package_version_resolved)

        runtime_env = context.project.runtime_environment
        runtime_env.operating_system.name = "rhel"
        runtime_env.operating_system.version = "8"
        runtime_env.python_version = "3.8"

        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        pv_tuple = ("daiquiri", "2.0.0", "https://pypi.org/simple")

        if pipeline_run:
            context.graph.should_receive("python_package_version_exists").with_args(
                *pv_tuple, solver_name="solver-rhel-8-py38"
            ).and_return(True).once()
            context.graph.should_receive("is_python_package_index_enabled").with_args(
                "https://pypi.org/simple"
            ).and_return(True).once()

        unit = AddPackageStepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

        if pipeline_run:
            assert "daiquiri" in state.unresolved_dependencies
            assert set(state.unresolved_dependencies.get("daiquiri").values()) == {pv_tuple}
            pv = context.get_package_version(pv_tuple, graceful=True)
            assert pv is not None
            assert pv.to_tuple() == pv_tuple
            assert pv.develop is True
        else:
            assert "daiquiri" not in state.unresolved_dependencies
            assert state.unresolved_dependencies.get("daiquiri") is None

    def test_run_package_version_not_solved(self, context: Context, state: State) -> None:
        """Test running the prescription based on the dependency not solved."""
        prescription_str = """
name: AddPackageStep
type: step.AddPackage
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: "~=1.19.0"
  state:
    package_version_from:
    - name: scikit-learn
      version: "<=0.25.0"
      develop: false
      index_url: "https://pypi.org/simple"
    resolved_dependencies:
    - name: click
      version: "==8.0.0"
      develop: false
      index_url: https://pypi.org/simple
run:
  package_version:
    name: daiquiri
    locked_version: ==2.0.0
    index_url: https://pypi.org/simple
    develop: true
  stack_info:
    - type: INFO
      message: "Hello, Thoth!"
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA(prescription)
        AddPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        package_version_resolved = PackageVersion(
            name="click",
            version="==8.0.0",
            index=pypi,
            develop=False,
        )

        assert "dauiqiri" not in state.resolved_dependencies

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        state.add_resolved_dependency(package_version_resolved.to_tuple())
        context.register_package_version(package_version_resolved)

        runtime_env = context.project.runtime_environment
        runtime_env.operating_system.name = "rhel"
        runtime_env.operating_system.version = "8"
        runtime_env.python_version = "3.8"

        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        pv_tuple = ("daiquiri", "2.0.0", "https://pypi.org/simple")

        context.graph.should_receive("python_package_version_exists").with_args(
            *pv_tuple, solver_name="solver-rhel-8-py38"
        ).and_return(False).once()

        unit = AddPackageStepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

        assert "daiquiri" not in state.unresolved_dependencies
        assert state.unresolved_dependencies.get("daiquiri") is None

    def test_run_package_version_index_url_not_enabled(self, context: Context, state: State) -> None:
        """Test running the prescription based on the dependency introduced when index_url is not enabled."""
        prescription_str = """
name: AddPackageStep
type: step.AddPackage
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: "~=1.19.0"
  state:
    package_version_from:
    - name: scikit-learn
      version: "<=0.25.0"
      develop: false
      index_url: "https://pypi.org/simple"
    resolved_dependencies:
    - name: click
      version: "==8.0.0"
      develop: false
      index_url: https://pypi.org/simple
run:
  package_version:
    name: daiquiri
    locked_version: ==2.0.0
    index_url: https://pypi.org/simple
    develop: true
  stack_info:
    - type: INFO
      message: "Hello, Thoth!"
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA(prescription)
        AddPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        package_version_resolved = PackageVersion(
            name="click",
            version="==8.0.0",
            index=pypi,
            develop=False,
        )

        assert "dauiqiri" not in state.resolved_dependencies

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        state.add_resolved_dependency(package_version_resolved.to_tuple())
        context.register_package_version(package_version_resolved)

        runtime_env = context.project.runtime_environment
        runtime_env.operating_system.name = "rhel"
        runtime_env.operating_system.version = "8"
        runtime_env.python_version = "3.8"

        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        pv_tuple = ("daiquiri", "2.0.0", "https://pypi.org/simple")

        context.graph.should_receive("python_package_version_exists").with_args(
            *pv_tuple, solver_name="solver-rhel-8-py38"
        ).and_return(True).once()
        context.graph.should_receive("is_python_package_index_enabled").with_args("https://pypi.org/simple").and_return(
            False
        ).once()

        unit = AddPackageStepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

        assert "daiquiri" not in state.unresolved_dependencies
        assert state.unresolved_dependencies.get("daiquiri") is None

    def test_run_package_version_index_url_not_known(self, context: Context, state: State) -> None:
        """Test running the prescription based on the dependency introduced when index_url is not known."""
        prescription_str = """
name: AddPackageStep
type: step.AddPackage
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: "~=1.19.0"
  state:
    package_version_from:
    - name: scikit-learn
      version: "<=0.25.0"
      develop: false
      index_url: "https://pypi.org/simple"
    resolved_dependencies:
    - name: click
      version: "==8.0.0"
      develop: false
      index_url: https://pypi.org/simple
run:
  package_version:
    name: daiquiri
    locked_version: ==2.0.0
    index_url: https://pypi.org/simple
    develop: true
  stack_info:
    - type: INFO
      message: "Hello, Thoth!"
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA(prescription)
        AddPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        package_version_resolved = PackageVersion(
            name="click",
            version="==8.0.0",
            index=pypi,
            develop=False,
        )

        assert "dauiqiri" not in state.resolved_dependencies

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        state.add_resolved_dependency(package_version_resolved.to_tuple())
        context.register_package_version(package_version_resolved)

        runtime_env = context.project.runtime_environment
        runtime_env.operating_system.name = "rhel"
        runtime_env.operating_system.version = "8"
        runtime_env.python_version = "3.8"

        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        pv_tuple = ("daiquiri", "2.0.0", "https://pypi.org/simple")

        context.graph.should_receive("python_package_version_exists").with_args(
            *pv_tuple, solver_name="solver-rhel-8-py38"
        ).and_return(True).once()
        context.graph.should_receive("is_python_package_index_enabled").with_args("https://pypi.org/simple").and_raise(
            NotFoundError
        ).once()

        unit = AddPackageStepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

        assert "daiquiri" not in state.unresolved_dependencies
        assert state.unresolved_dependencies.get("daiquiri") is None

    def test_run_package_version_already_resolved(self, context: Context, state: State) -> None:
        """Test running the prescription when the given package is already in the resolved state."""
        prescription_str = """
name: AddPackageStep
type: step.AddPackage
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: "~=1.19.0"
  state:
    package_version_from:
    - name: scikit-learn
      version: "<=0.25.0"
      develop: false
      index_url: "https://pypi.org/simple"
    resolved_dependencies:
    - name: click
      version: "==8.0.0"
      develop: false
      index_url: https://pypi.org/simple
run:
  package_version:
    name: daiquiri
    locked_version: ==2.0.0
    index_url: https://pypi.org/simple
    develop: true
  stack_info:
    - type: INFO
      message: "Hello, Thoth!"
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA(prescription)
        AddPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        package_version_resolved = PackageVersion(
            name="click",
            version="==8.0.0",
            index=pypi,
            develop=False,
        )

        pv_daiquiri = PackageVersion(
            name="daiquiri",
            version="==2.0.0",
            index=pypi,
            develop=False,
        )

        pv_tuple = ("daiquiri", "2.0.0", "https://pypi.org/simple")

        state.add_resolved_dependency(pv_tuple)
        context.register_package_version(pv_daiquiri)

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        state.add_resolved_dependency(package_version_resolved.to_tuple())
        context.register_package_version(package_version_resolved)

        runtime_env = context.project.runtime_environment
        runtime_env.operating_system.name = "rhel"
        runtime_env.operating_system.version = "8"
        runtime_env.python_version = "3.8"

        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        unit = AddPackageStepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

        assert "daiquiri" not in state.unresolved_dependencies
        assert "daiquiri" in state.resolved_dependencies

    def test_run_package_version_already_resolved_same_name(self, context: Context, state: State) -> None:
        """Test running the prescription when the given package is already in the resolved state (same name)."""
        prescription_str = """
name: AddPackageStep
type: step.AddPackage
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_version:
    name: numpy
    version: "~=1.19.0"
  state:
    package_version_from:
    - name: scikit-learn
      version: "<=0.25.0"
      develop: false
      index_url: "https://pypi.org/simple"
    resolved_dependencies:
    - name: click
      version: "==8.0.0"
      develop: false
      index_url: https://pypi.org/simple
run:
  package_version:
    name: daiquiri
    locked_version: ==2.0.0
    index_url: https://pypi.org/simple
    develop: true
  stack_info:
    - type: INFO
      message: "Hello, Thoth!"
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA(prescription)
        AddPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        package_version_from = PackageVersion(
            name="scikit-learn",
            version="==0.24.2",
            index=pypi,
            develop=False,
        )

        package_version_resolved = PackageVersion(
            name="click",
            version="==8.0.0",
            index=pypi,
            develop=False,
        )

        pv_daiquiri = PackageVersion(
            name="daiquiri",
            version="==2.5.0",
            index=pypi,
            develop=False,
        )

        pv_tuple = ("daiquiri", "2.5.0", "https://pypi.org/simple")

        state.add_resolved_dependency(pv_tuple)
        context.register_package_version(pv_daiquiri)

        state.add_resolved_dependency(package_version_from.to_tuple())
        context.register_package_version(package_version_from)

        state.add_resolved_dependency(package_version_resolved.to_tuple())
        context.register_package_version(package_version_resolved)

        runtime_env = context.project.runtime_environment
        runtime_env.operating_system.name = "rhel"
        runtime_env.operating_system.version = "8"
        runtime_env.python_version = "3.8"

        context.register_package_tuple(
            package_version.to_tuple(),
            dependent_tuple=package_version_from.to_tuple(),
            develop=False,
            extras=None,
            os_name=runtime_env.operating_system.name,
            os_version=runtime_env.operating_system.version,
            python_version=runtime_env.python_version,
        )

        unit = AddPackageStepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version) is None

        assert "daiquiri" not in state.unresolved_dependencies
        assert "daiquiri" in state.resolved_dependencies
        assert set(state.resolved_dependencies["daiquiri"]) == set(pv_tuple)

    def test_run_package_version_add_package_multi(self, context: Context, state: State) -> None:
        """Test running the unit multiple times with stack info and justification reported once."""
        prescription_str = """
name: AddPackageStep
type: step.AddPackage
should_include:
  adviser_pipeline: true
match:
  package_version:
    name: numpy
run:
  package_version:
    name: daiquiri
    locked_version: ==2.0.0
    index_url: https://pypi.org/simple
    develop: true
  stack_info:
  - type: INFO
    message: "Hello, Thoth!"
    link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA(prescription)
        AddPackageStepPrescription.set_prescription(prescription)

        pypi = Source("https://pypi.org/simple")
        package_version_np = PackageVersion(
            name="numpy",
            version="==1.19.1",
            index=pypi,
            develop=False,
        )

        context.stack_info.clear()

        context.graph.should_receive("python_package_version_exists").with_args(
            "daiquiri", "2.0.0", "https://pypi.org/simple", solver_name="solver-rhel-8-py38"
        ).and_return(True).once()
        context.graph.should_receive("is_python_package_index_enabled").with_args("https://pypi.org/simple").and_return(
            True
        ).once()

        state.resolved_dependencies.clear()

        runtime_env = context.project.runtime_environment
        runtime_env.operating_system.name = "rhel"
        runtime_env.operating_system.version = "8"
        runtime_env.python_version = "3.8"

        unit = AddPackageStepPrescription()
        unit.pre_run()
        with unit.assigned_context(context):
            assert unit.run(state, package_version_np) is None

        assert "daiquiri" in state.unresolved_dependencies
        assert set(state.unresolved_dependencies["daiquiri"].values()) == {
            ("daiquiri", "2.0.0", "https://pypi.org/simple")
        }

        assert context.stack_info == [
            {"type": "INFO", "message": "Hello, Thoth!", "link": "https://thoth-station.ninja"}
        ]

        package_version_np2 = PackageVersion(
            name="numpy",
            version="==2.0.0",
            index=pypi,
            develop=False,
        )

        context.graph.should_receive("python_package_version_exists").with_args(
            "daiquiri", "2.0.0", "https://pypi.org/simple", solver_name="solver-rhel-8-py38"
        ).and_return(True).once()
        context.graph.should_receive("is_python_package_index_enabled").with_args("https://pypi.org/simple").and_return(
            True
        ).once()

        state.resolved_dependencies.clear()

        runtime_env = context.project.runtime_environment
        runtime_env.operating_system.name = "rhel"
        runtime_env.operating_system.version = "8"
        runtime_env.python_version = "3.8"

        with unit.assigned_context(context):
            assert unit.run(state, package_version_np2) is None

        assert set(state.unresolved_dependencies["daiquiri"].values()) == {
            ("daiquiri", "2.0.0", "https://pypi.org/simple")
        }
        assert context.stack_info == [
            {"type": "INFO", "message": "Hello, Thoth!", "link": "https://thoth-station.ninja"}
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to check if recommendation is targeting a supported platform."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr

from thoth.common import get_justification_link as jl

from ..boot import Boot
from ..exceptions import NotAcceptable

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PlatformBoot(Boot):
    """A boot to check if a supported platform is used.

    We could check this based on the database entries, but as this will change rarely, we can
    hardcode supported platforms here.
    """

    _JUSTIFICATION_LINK = jl("platform")
    _SUPPORTED_PLATFORMS = frozenset({"linux-x86_64"})

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, always on unsupported platform."""
        if (
            builder_context.project.runtime_environment.platform not in cls._SUPPORTED_PLATFORMS
            and not builder_context.is_included(cls)
        ):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Check if GPU enabled recommendations should be done."""
        platform = self.context.project.runtime_environment.platform
        msg = f"Platform {platform!r} is not supported, possible platforms are: " + ", ".join(self._SUPPORTED_PLATFORMS)
        self.context.stack_info.append(
            {
                "type": "ERROR",
                "message": msg,
                "link": self._JUSTIFICATION_LINK,
            }
        )
        raise NotAcceptable(f"{msg} - see {self._JUSTIFICATION_LINK}")



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Routines for dependency monkey and its output handling."""

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Union

import attr

from .product import Product


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class DependencyMonkeyReport:
    """Report produced by a Dependency Monkey run."""

    skipped = attr.ib(type=int, default=0)
    _responses = attr.ib(type=List[Dict[str, Union[Dict[str, Any], str]]], default=attr.Factory(list))

    def add_response(self, response: str, product: Product) -> None:
        """Add a new response to response listing."""
        self._responses.append({"response": response, "product": product.to_dict()})

    def to_dict(self, *, verbose: bool = False) -> Dict[str, Any]:
        """Convert report to a dict representation suitable for serialization."""
        return {"skipped": self.skipped, "responses": self._responses}



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing predictor."""

import abc
import logging
from contextlib import contextmanager

import attr
from typing import Any
from typing import Tuple
from typing import Optional
from typing import Generator
from typing import TYPE_CHECKING

from .context import Context
from .report import Report
from .state import State
from .utils import should_keep_history

if TYPE_CHECKING:
    import matplotlib.figure


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Predictor:
    """A base class for implementing a predictor for resolver."""

    keep_history = attr.ib(type=bool, kw_only=True, default=None, converter=should_keep_history)

    _CONTEXT: Optional[Context] = None

    @classmethod
    def obtain_default_configuration(cls, config_option_name: str) -> Any:
        """Get the default value assigned to the given configuration option of a predictor instance."""
        return getattr(cls.__attrs_attrs__, config_option_name).default

    @classmethod
    @contextmanager
    def assigned_context(cls, context: Context) -> Generator[None, None, None]:
        """Assign context to predictor."""
        try:
            cls._CONTEXT = context
            yield
        finally:
            cls._CONTEXT = None

    @property
    def context(self) -> Context:
        """Get context in which the unit runs in."""
        if self._CONTEXT is None:
            raise ValueError("Requesting resolver context outside of resolver run")

        return self._CONTEXT

    def pre_run(self) -> None:
        """Pre-initialize the predictor.

        This method is called before any resolving with a freshly instantiated context. The default operation is a noop,
        but predictor can perform any initial setup in this method. This method should not raise any exception.
        """
        # noop

    @abc.abstractmethod
    def run(self) -> Tuple[State, Tuple[str, str, str]]:
        """Run the main method used to run the predictor."""
        raise NotImplementedError

    def post_run(self) -> None:
        """Post-run method run after the resolving has been done."""
        # noop

    def post_run_report(self, report: Report) -> None:
        """Post-run method run after the resolving has finished - this method is called only if resolving with a report.

        The default operation is a noop, but a predictor can perform any post-run operations in this method. This
        method should not raise any exception.
        """
        # noop

    def set_reward_signal(self, state: State, package_tuple: Tuple[str, str, str], reward: float) -> None:
        """Signalize the reward.

        @param state: (child) state for which the reward signal is triggered
        @param package_tuple: Python package that was added to the state causing the reward
        @param reward: set to nan if the given state was not accepted a special value
                       of inf notifies about a new final state
        """
        # noop

    def finalize_state(self, state_id: int) -> None:  # noqa: D401
        """Finalizer called when the given state is about to be destructed by garbage collector.

        Method suitable if predictor keeps internal state for states. Note that this method is not
        called for remaining states if the resolver terminates.

        @param state_id: id of state that is about to be finalized
        """
        # noop

    def plot(self) -> "matplotlib.figure.Figure":
        """Plot information about predictor."""
        _LOGGER.error(
            "Cannot plot predictor history as plotting is not implemented for predictor %r, error is not fatal",
            self.__class__.__name__,
        )

    @staticmethod
    def _make_patch_spines_invisible(ax: Any) -> None:
        """Make spines invisible."""
        ax.set_frame_on(True)
        ax.patch.set_visible(False)
        for sp in ax.spines.values():
            sp.set_visible(False)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test scoring (penalization) based on a CVE."""

from flexmock import flexmock
import pytest

from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.steps import CvePenalizationStep
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase

from ..base import AdviserUnitTestCase


class TestCvePenalizationStep(AdviserUnitTestCase):
    """Test scoring (penalization) based on a CVE."""

    UNIT_TESTED = CvePenalizationStep

    _CASE_CANDIDATES = [
        ("flask", "0.12.0", "https://pypi.org/simple"),
        ("click", "2.0", "https://pypi.org/simple"),
        ("pyyaml", "3.12", "https://pypi.org/simple"),
    ]

    _FLASK_CVE = {
        "details": "flask version Before 0.12.3 contains a CWE-20: Improper Input Validation "
        "vulnerability in flask that can result in Large amount of memory usage "
        "possibly leading to denial of service.",
        "cve_id": "CVE-ID",
        "aggregated_at": "2021-06-02T08:23:17.11783Z",
    }

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.LATEST
        self.verify_multiple_should_include(builder_context)

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.SECURITY])
    def test_no_include(self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType):
        """Check not including the given pipeline unit for recommendation type SECURITY."""
        builder_context.recommendation_type = recommendation_type
        builder_context.decision_type = None
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    @pytest.mark.parametrize(
        "recommendation_type",
        [
            RecommendationType.LATEST,
            RecommendationType.TESTING,
            RecommendationType.STABLE,
            RecommendationType.PERFORMANCE,
        ],
    )
    def test_should_include_recommendation_type(
        self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType
    ):
        """Check inclusion for various recommendation types."""
        builder_context.recommendation_type = recommendation_type
        builder_context.decision_type = None
        if recommendation_type in (RecommendationType.LATEST, RecommendationType.TESTING):
            assert list(self.UNIT_TESTED.should_include(builder_context)) == [{"cve_penalization": 0.0}]
        else:
            assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.PERFORMANCE, RecommendationType.STABLE])
    def test_cve_penalization(self, recommendation_type: RecommendationType) -> None:
        """Make sure a CVE affects stack score."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="0.12.0"
        ).and_return([self._FLASK_CVE]).once()

        package_version = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(graph=GraphDatabase(), recommendation_type=recommendation_type)
        with CvePenalizationStep.assigned_context(context):
            step = CvePenalizationStep()
            result = step.run(None, package_version)

        assert result is not None
        assert isinstance(result, tuple) and len(result) == 2
        assert isinstance(result[0], float)
        assert result[0] == 1 * CvePenalizationStep.CONFIGURATION_DEFAULT["cve_penalization"]
        assert isinstance(result[1], list)
        assert result[1] == [
            {
                "link": "https://thoth-station.ninja/j/cve",
                "message": "Package  ('flask', '0.12.0', 'https://pypi.org/simple') has a CVE 'CVE-ID'",
                "advisory": "flask version Before 0.12.3 contains a CWE-20: Improper Input Validation "
                "vulnerability in flask that can result in Large amount of memory usage "
                "possibly leading to denial of service.",
                "package_name": "flask",
                "type": "WARNING",
            }
        ]
        assert self.verify_justification_schema(result[1])

    @pytest.mark.parametrize("recommendation_type", [RecommendationType.LATEST, RecommendationType.TESTING])
    def test_cve_no_penalization(self, recommendation_type: RecommendationType) -> None:
        """Make sure score penalization is not done for testing and latest recommendation types."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="0.12.0"
        ).and_return([self._FLASK_CVE]).once()

        package_version = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(graph=GraphDatabase(), recommendation_type=recommendation_type, stack_info=[])
        with CvePenalizationStep.assigned_context(context):
            step = CvePenalizationStep()
            result = step.run(None, package_version)

        assert result is not None
        assert isinstance(result, tuple) and len(result) == 2
        assert isinstance(result[0], float)
        assert result[0] == 0.0
        assert isinstance(result[1], list)
        assert result[1] == [
            {
                "link": "https://thoth-station.ninja/j/cve",
                "message": "Package  ('flask', '0.12.0', 'https://pypi.org/simple') has a CVE 'CVE-ID'",
                "advisory": "flask version Before 0.12.3 contains a CWE-20: Improper Input Validation "
                "vulnerability in flask that can result in Large amount of memory usage "
                "possibly leading to denial of service.",
                "package_name": "flask",
                "type": "WARNING",
            }
        ]
        assert self.verify_justification_schema(result[1])

    @pytest.mark.parametrize("recommendation_type", RecommendationType.__members__.values())
    def test_no_cve_record(self, recommendation_type: RecommendationType) -> None:
        """Make sure no CVEs do not affect CVE scoring."""
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="0.12.0"
        ).and_return([]).once()

        package_version = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(graph=GraphDatabase(), recommendation_type=recommendation_type)
        with CvePenalizationStep.assigned_context(context):
            step = CvePenalizationStep()
            result = step.run(None, package_version)

        assert isinstance(result, tuple)
        assert len(result) == 2
        assert result[0] == 0.0
        assert result[1] == [
            {
                "link": "https://thoth-station.ninja/j/no_cve",
                "message": "No known CVE known for 'flask' in version '0.12.0'",
                "package_name": "flask",
                "type": "INFO",
            }
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test filtering based on a CVE."""

import pytest

from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import CveSieve
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from thoth.python import Source

from ..base import AdviserUnitTestCase


class TestCveSieve(AdviserUnitTestCase):
    """Test filtering based on a CVE."""

    UNIT_TESTED = CveSieve

    _CASE_CANDIDATES = [
        ("flask", "0.12.0", "https://pypi.org/simple"),
        ("click", "2.0", "https://pypi.org/simple"),
        ("pyyaml", "3.12", "https://pypi.org/simple"),
    ]

    _FLASK_CVE = {
        "details": "flask version Before 0.12.3 contains a CWE-20: Improper Input Validation "
        "vulnerability in flask that can result in Large amount of memory usage "
        "possibly leading to denial of service.",
        "cve_id": "CVE-ID",
        "aggregated_at": "2021-06-02T08:23:17.11783Z",
    }

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.recommendation_type = RecommendationType.SECURITY
        self.verify_multiple_should_include(builder_context)

    @pytest.mark.parametrize(
        "recommendation_type",
        [
            RecommendationType.LATEST,
            RecommendationType.STABLE,
            RecommendationType.PERFORMANCE,
            RecommendationType.TESTING,
        ],
    )
    def test_should_include_recommendation_type(
        self, builder_context: PipelineBuilderContext, recommendation_type: RecommendationType
    ):
        """Check not including the given pipeline unit for recommendations other than SECURITY."""
        builder_context.recommendation_type = recommendation_type
        builder_context.decision_type = None
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_cve_sieve(self, context: Context) -> None:
        """Make sure a CVE filters out packages."""
        context.graph.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="0.12.0"
        ).and_return([self._FLASK_CVE])

        context.graph.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="1.0.0"
        ).and_return([])

        pv1 = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        pv2 = PackageVersion(
            name="flask",
            version="==1.0.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        assert not context.stack_info

        context.recommendation_type = RecommendationType.SECURITY

        with self.UNIT_TESTED.assigned_context(context):
            unit = self.UNIT_TESTED()
            result = list(unit.run((pv for pv in (pv1, pv2))))

        assert result is not None
        assert result == [pv2]
        assert context.stack_info == [
            {
                "link": jl("cve"),
                "message": "Skipping including package ('flask', '0.12.0', "
                "'https://pypi.org/simple') as a CVE 'CVE-ID' was found",
                "type": "WARNING",
            }
        ]
        assert self.verify_justification_schema(context.stack_info)

    def test_cve_sieve_allow_cve(self, context: Context) -> None:
        """Make sure a CVE filtering allows allow-cve configuration."""
        context.graph.should_receive("get_python_cve_records_all").with_args(
            package_name="flask", package_version="0.12.0"
        ).and_return([self._FLASK_CVE])

        pv = PackageVersion(
            name="flask",
            version="==0.12.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        assert not context.stack_info

        context.recommendation_type = RecommendationType.SECURITY

        context.labels["allow-cve"] = self._FLASK_CVE["cve_id"]
        with self.UNIT_TESTED.assigned_context(context):
            unit = self.UNIT_TESTED()
            unit.pre_run()
            result = list(unit.run((p for p in (pv,))))

        assert result == [pv]
        assert context.stack_info == []



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of a prescription abstraction."""

import logging
import os
import pickle
import yaml
from itertools import chain

from collections import OrderedDict
from collections import deque
from typing import Any
from typing import List
from typing import Tuple
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING

import attr

from ...exceptions import InternalError
from ...exceptions import PrescriptionDuplicateUnitNameError
from ...exceptions import PrescriptionSchemaError
from .add_package_step import AddPackageStepPrescription
from .boot import BootPrescription
from .gh_release_notes import GHReleaseNotesWrapPrescription
from .group import GroupStepPrescription
from .pseudonym import PseudonymPrescription
from .schema import PRESCRIPTION_SCHEMA
from .sieve import SievePrescription
from .skip_package_sieve import SkipPackageSievePrescription
from .skip_package_step import SkipPackageStepPrescription
from .step import StepPrescription
from .stride import StridePrescription
from .wrap import WrapPrescription

if TYPE_CHECKING:
    from thoth.adviser.unit_types import UnitType  # noqa: F401
    from thoth.adviser.unit_types import BootType  # noqa: F401
    from thoth.adviser.unit_types import PseudonymType  # noqa: F401
    from thoth.adviser.unit_types import SieveType  # noqa: F401
    from thoth.adviser.unit_types import StepType  # noqa: F401
    from thoth.adviser.unit_types import StrideType  # noqa: F401
    from thoth.adviser.unit_types import WrapType  # noqa: F401


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Prescription:
    """Dynamically create pipeline units based on inscription."""

    _PRESCRIPTION_METADATA_FILE = "_prescription_metadata.yaml"
    _PRESCRIPTION_DEFAULT_NAME = "UNKNOWN"
    _PRESCRIPTION_DEFAULT_RELEASE = "UNKNOWN"
    _VALIDATE_PRESCRIPTION_SCHEMA = bool(int(os.getenv("THOTH_ADVISER_VALIDATE_PRESCRIPTION_SCHEMA", 1)))

    prescriptions = attr.ib(type=List[Tuple[str, str]], kw_only=True, default=attr.Factory(list))

    boots_dict = attr.ib(type=Dict[str, Dict[str, Any]], kw_only=True, default=attr.Factory(OrderedDict))
    pseudonyms_dict = attr.ib(type=Dict[str, Dict[str, Any]], kw_only=True, default=attr.Factory(OrderedDict))
    sieves_dict = attr.ib(type=Dict[str, Dict[str, Any]], kw_only=True, default=attr.Factory(OrderedDict))
    steps_dict = attr.ib(type=Dict[str, Dict[str, Any]], kw_only=True, default=attr.Factory(OrderedDict))
    strides_dict = attr.ib(type=Dict[str, Dict[str, Any]], kw_only=True, default=attr.Factory(OrderedDict))
    wraps_dict = attr.ib(type=Dict[str, Dict[str, Any]], kw_only=True, default=attr.Factory(OrderedDict))

    @property
    def units(self) -> Generator[Dict[str, Any], None, None]:
        """Iterate over units."""
        yield from chain(
            self.boots_dict.values(),
            self.pseudonyms_dict.values(),
            self.sieves_dict.values(),
            self.steps_dict.values(),
            self.strides_dict.values(),
            self.wraps_dict.values(),
        )

    @staticmethod
    def _validate_run_base(unit: Dict[str, Any]) -> bool:
        """Validate run for a pipeline unit."""
        if unit.get("run", {}).get("eager_stop_pipeline") and unit.get("run", {}).get("not_acceptable"):
            _LOGGER.error(
                "Error in unit %s (%s): Using 'eager_stop_pipeline' and 'not_acceptable' at the same "
                "time has undefined behavior",
                unit["name"],
                unit["type"],
            )

        return False

    @classmethod
    def validate(cls, prescriptions: str) -> "Prescription":
        """Validate the given prescription."""
        _LOGGER.debug("Validating prescriptions schema")

        prescription_instance = cls.load(prescriptions)

        # Verify semantics of prescription.
        any_error = False

        for unit in prescription_instance.units:
            any_error = any_error or cls._validate_run_base(unit)

        for pseudonym in prescription_instance.pseudonyms_dict.values():
            if pseudonym["run"]["yield"].get("yield_matched_version") and pseudonym["run"]["yield"][
                "package_version"
            ].get("locked_version"):
                _LOGGER.error(
                    "Error in unit %s (%s): Using 'yield_matched_version' together "
                    "with 'locked_version' leads to undefined behavior",
                    pseudonym["name"],
                    pseudonym["type"],
                )
                any_error = True

        for step in prescription_instance.steps_dict.values():
            step_run = step["run"]
            msg = f"Error in unit {step['name']} ({step['type']})"

            if step_run.get("eager_stop_pipeline") and step_run.get("justification"):
                _LOGGER.error(
                    "%s: Using 'eager_stop_pipeline' together with 'justification' leads to undefined behavior", msg
                )
                any_error = True

            if step_run.get("not_acceptable") and step_run.get("justification"):
                _LOGGER.error(
                    "%s: Using 'not_acceptable' together with 'justification' leads to undefined behavior", msg
                )

            if step_run.get("eager_stop_pipeline") and step_run.get("score") is not None:
                _LOGGER.error("%s: Using 'eager_stop_pipeline' together with 'score' leads to undefined behavior", msg)
                any_error = True

            if step_run.get("not_acceptable") and step_run.get("score") is not None:
                _LOGGER.error("%s: Using 'not_acceptable' together with 'score' leads to undefined behavior", msg)
                any_error = True

        for wrap in prescription_instance.wraps_dict.values():
            wrap_run = wrap["run"]
            msg = f"Error in unit {wrap['name']} ({wrap['type']})"

            if wrap_run.get("eager_stop_pipeline") and wrap_run.get("justification"):
                _LOGGER.error(
                    "%s: Using 'eager_stop_pipeline' together with 'justification' leads to undefined behavior", msg
                )
                any_error = True

            if wrap_run.get("not_acceptable") and wrap_run.get("justification"):
                _LOGGER.error(
                    "%s: Using 'not_acceptable' together with 'justification' leads to undefined behavior", msg
                )

        if any_error:
            raise PrescriptionSchemaError("Failed to validate prescription units, see logs reported for more info")

        # Drop any metadata associated to save space.
        for unit in prescription_instance.units:
            unit.pop("metadata", None)

        return prescription_instance

    @classmethod
    def from_dict(
        cls,
        prescription: Dict[str, Any],
        *,
        prescription_instance: Optional["Prescription"] = None,
        prescription_name: str,
        prescription_release: str,
    ) -> "Prescription":
        """Instantiate prescription from a dictionary representation.

        If an instance is provided, a safe merge will be performed.
        """
        try:
            PRESCRIPTION_SCHEMA(prescription)
        except Exception as exc:
            _LOGGER.exception(
                "Failed to validate schema for prescription: %s",
                str(exc),
            )
            raise PrescriptionSchemaError(str(exc))

        boots_dict = prescription_instance.boots_dict if prescription_instance is not None else OrderedDict()
        for boot_spec in prescription["units"].get("boots") or []:
            name = f"{prescription_name}.{boot_spec['name']}"
            boot_spec["name"] = name
            if name in boots_dict:
                raise PrescriptionDuplicateUnitNameError(f"Boot with name {name!r} is already present")
            boots_dict[boot_spec["name"]] = boot_spec

        pseudonyms_dict = prescription_instance.pseudonyms_dict if prescription_instance else OrderedDict()
        for pseudonym_spec in prescription["units"].get("pseudonyms") or []:
            name = f"{prescription_name}.{pseudonym_spec['name']}"
            pseudonym_spec["name"] = name
            if name in pseudonyms_dict:
                raise PrescriptionDuplicateUnitNameError(f"Pseudonym with name {name!r} is already present")
            pseudonyms_dict[pseudonym_spec["name"]] = pseudonym_spec

        sieves_dict = prescription_instance.sieves_dict if prescription_instance else OrderedDict()
        for sieve_spec in prescription["units"].get("sieves") or []:
            name = f"{prescription_name}.{sieve_spec['name']}"
            sieve_spec["name"] = name
            if name in sieves_dict:
                raise PrescriptionDuplicateUnitNameError(f"Sieve with name {name!r} is already present")
            sieves_dict[sieve_spec["name"]] = sieve_spec

        steps_dict = prescription_instance.steps_dict if prescription_instance else OrderedDict()
        for step_spec in prescription["units"].get("steps") or []:
            name = f"{prescription_name}.{step_spec['name']}"
            step_spec["name"] = name
            if name in steps_dict:
                raise PrescriptionDuplicateUnitNameError(f"Step with name {name!r} is already present")
            steps_dict[step_spec["name"]] = step_spec

        strides_dict = prescription_instance.strides_dict if prescription_instance else OrderedDict()
        for stride_spec in prescription["units"].get("strides") or []:
            name = f"{prescription_name}.{stride_spec['name']}"
            stride_spec["name"] = name
            if name in strides_dict:
                raise PrescriptionDuplicateUnitNameError(f"Stride with name {name!r} is already present")
            strides_dict[stride_spec["name"]] = stride_spec

        wraps_dict = prescription_instance.wraps_dict if prescription_instance else OrderedDict()
        for wrap_spec in prescription["units"].get("wraps") or []:
            name = f"{prescription_name}.{wrap_spec['name']}"
            wrap_spec["name"] = name
            if name in wraps_dict:
                raise PrescriptionDuplicateUnitNameError(f"Wrap with name {name!r} is already present")
            wraps_dict[wrap_spec["name"]] = wrap_spec

        if prescription_instance:
            # Adjust release info at the end once successful.
            prescription_meta = (prescription_name, prescription_release)
            if prescription_meta not in prescription_instance.prescriptions:
                prescription_instance.prescriptions.append(prescription_meta)
            return prescription_instance

        return cls(
            boots_dict=boots_dict,
            pseudonyms_dict=pseudonyms_dict,
            sieves_dict=sieves_dict,
            steps_dict=steps_dict,
            strides_dict=strides_dict,
            wraps_dict=wraps_dict,
            prescriptions=[(prescription_name, prescription_release)],
        )

    def is_empty(self) -> bool:
        """Check if no prescription units are loaded."""
        return (
            not self.boots_dict
            and not self.pseudonyms_dict
            and not self.sieves_dict
            and not self.steps_dict
            and not self.strides_dict
            and not self.wraps_dict
        )

    @classmethod
    def load(cls, *prescriptions: str) -> "Prescription":
        """Load prescription from files or from their YAML representation."""
        queue = deque([(p, cls._PRESCRIPTION_DEFAULT_NAME, cls._PRESCRIPTION_DEFAULT_RELEASE) for p in prescriptions])
        prescription_instance = Prescription()

        while queue:
            prescription, prescription_name, prescription_release = queue.popleft()

            if os.path.isfile(prescription):
                if prescription.endswith((".yaml", ".yml")):
                    _LOGGER.debug("Loading prescriptions from %r", prescription)
                    with open(prescription, "r") as config_file:
                        try:
                            prescription_instance = cls.from_dict(
                                yaml.load(config_file, Loader=yaml.CLoader),
                                prescription_instance=prescription_instance,
                                prescription_name=prescription_name,
                                prescription_release=prescription_release,
                            )
                        except Exception:
                            _LOGGER.error("Failed to load prescription from %r", prescription)
                            raise
                elif prescription.endswith(".pickle"):
                    _LOGGER.debug("Loading prescriptions from %r", prescription)
                    with open(prescription, "rb") as fp:
                        prescription_instance = pickle.load(fp)

                    for prescription_info in prescription_instance.prescriptions:
                        _LOGGER.info(
                            "Loaded prescriptions %r in version %r", prescription_info[0], prescription_info[1]
                        )
                else:
                    _LOGGER.debug("Skipping file %r: not a YAML nor Pickle file", prescription)
                    continue

            elif os.path.isdir(prescription):
                prescription_metadata_file_path = os.path.join(prescription, cls._PRESCRIPTION_METADATA_FILE)
                if os.path.isfile(prescription_metadata_file_path):
                    with open(prescription_metadata_file_path) as f:
                        _LOGGER.debug("Loading prescription metadata file %r", prescription_metadata_file_path)
                        metadata = yaml.safe_load(f)

                        prescription_name = metadata.get("prescription", {}).get("name", cls._PRESCRIPTION_DEFAULT_NAME)
                        if prescription_name == cls._PRESCRIPTION_DEFAULT_NAME:
                            _LOGGER.warning("No prescription name parsed from %r", prescription_metadata_file_path)

                        prescription_release = metadata.get("prescription", {}).get(
                            "release", cls._PRESCRIPTION_DEFAULT_RELEASE
                        )
                        if prescription_release == cls._PRESCRIPTION_DEFAULT_RELEASE:
                            _LOGGER.warning("No prescription release parsed from %r", prescription_metadata_file_path)

                        _LOGGER.info("Using prescriptions %r release %r", prescription_name, prescription_release)

                for file_ in os.listdir(prescription):
                    if file_ == cls._PRESCRIPTION_METADATA_FILE:
                        continue

                    if file_.startswith("."):
                        _LOGGER.debug("Skipping hidden file or directory %r", file_)
                        continue

                    queue.append((os.path.join(prescription, file_), prescription_name, prescription_release))
            else:
                # Passed using string.
                prescription_instance = cls.from_dict(
                    yaml.safe_load(prescription),
                    prescription_instance=prescription_instance,
                    prescription_name=prescription_name,
                    prescription_release=prescription_release,
                )

        if prescription_instance.is_empty():
            raise InternalError("No prescription loaded")

        return prescription_instance

    @staticmethod
    def _iter_units(unit_class: Type["UnitType"], units: Dict[str, Any]) -> Generator[Type["UnitType"], None, None]:
        """Iterate over units registered."""
        for prescription in units.values():
            unit_class.set_prescription(prescription)

            yield unit_class

    def iter_boot_units(self) -> Generator[Type["BootType"], None, None]:
        """Iterate over prescription boot units registered in the prescription supplied."""
        return self._iter_units(BootPrescription, self.boots_dict)

    def iter_pseudonym_units(self) -> Generator[Type["PseudonymType"], None, None]:
        """Iterate over prescription pseudonym units registered in the prescription supplied."""
        return self._iter_units(PseudonymPrescription, self.pseudonyms_dict)

    def iter_sieve_units(self) -> Generator[Type["SieveType"], None, None]:
        """Iterate over prescription sieve units registered in the prescription supplied."""
        for prescription in self.sieves_dict.values():
            if prescription["type"] == "sieve":
                SievePrescription.set_prescription(prescription)
                yield SievePrescription
            elif prescription["type"] == "sieve.SkipPackage":
                SkipPackageSievePrescription.set_prescription(prescription)
                yield SkipPackageSievePrescription
            else:
                raise ValueError(f"Unknown sieve pipeline unit type: {prescription['type']!r}")

    def iter_step_units(self) -> Generator[Type["StepType"], None, None]:
        """Iterate over prescription step units registered in the prescription supplied."""
        for prescription in self.steps_dict.values():
            if prescription["type"] == "step":
                StepPrescription.set_prescription(prescription)
                yield StepPrescription
            elif prescription["type"] == "step.SkipPackage":
                SkipPackageStepPrescription.set_prescription(prescription)
                yield SkipPackageStepPrescription
            elif prescription["type"] == "step.AddPackage":
                AddPackageStepPrescription.set_prescription(prescription)
                yield AddPackageStepPrescription
            elif prescription["type"] == "step.Group":
                GroupStepPrescription.set_prescription(prescription)
                yield GroupStepPrescription
            else:
                raise ValueError(f"Unknown step pipeline unit type: {prescription['type']!r}")

    def iter_stride_units(self) -> Generator[Type["StrideType"], None, None]:
        """Iterate over prescription stride units registered in the prescription supplied."""
        return self._iter_units(StridePrescription, self.strides_dict)

    def iter_wrap_units(self) -> Generator[Type["WrapType"], None, None]:
        """Iterate over prescription stride units registered in the prescription supplied."""
        for prescription in self.wraps_dict.values():
            if prescription["type"] == "wrap":
                WrapPrescription.set_prescription(prescription)
                yield WrapPrescription
            elif prescription["type"] == "wrap.GHReleaseNotes":
                GHReleaseNotesWrapPrescription.set_prescription(prescription)
                yield GHReleaseNotesWrapPrescription
            else:
                raise ValueError(f"Unknown wrap pipeline unit type: {prescription['type']!r}")



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing wrap units."""

import attr

from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

from thoth.adviser.state import State
from voluptuous import Any as SchemaAny
from voluptuous import Optional
from voluptuous import Required
from voluptuous import Schema

from .unit import UnitPrescription
from .schema import PRESCRIPTION_WRAP_MATCH_ENTRY_SCHEMA
from .schema import PRESCRIPTION_WRAP_RUN_SCHEMA

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext


@attr.s(slots=True)
class WrapPrescription(UnitPrescription):
    """Wrap base class implementation."""

    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): SchemaAny(str, None),
            Optional("match"): PRESCRIPTION_WRAP_MATCH_ENTRY_SCHEMA,
            Required("run"): PRESCRIPTION_WRAP_RUN_SCHEMA,
            Required("prescription"): Schema({"run": bool}),
        }
    )

    @staticmethod
    def is_wrap_unit_type() -> bool:
        """Check if this unit is of type wrap."""
        return True

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if the given pipeline unit should be included in the given pipeline configuration."""
        if cls._should_include_base(builder_context):
            prescription: Dict[str, Any] = cls._PRESCRIPTION  # type: ignore
            yield from cls._yield_should_include_with_state(prescription)
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Prepare this pipeline unit before run."""
        self._prepare_justification_link(self.run_prescription.get("justification", []))
        super().pre_run()

    def run(self, state: State) -> None:
        """Run main entry-point for wrap units to filter and score packages."""
        if not self._run_state(state):
            return None

        prescription_conf = self._configuration["prescription"]

        if not prescription_conf["run"]:
            justification = self.run_prescription.get("justification")
            if justification:
                state.add_justification(justification)

            advised_manifest_changes = self.run_prescription.get("advised_manifest_changes")
            if advised_manifest_changes:
                state.advised_manifest_changes.append(advised_manifest_changes)

        try:
            self._run_base()
        finally:
            prescription_conf["run"] = True



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test pipeline builder used for building pipeline configuration."""

import os
from typing import Union
from typing import Dict

from flexmock import flexmock
import pytest
from pathlib import Path
import json
import yaml

from thoth.adviser.boot import Boot
from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilder
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.pseudonym import Pseudonym
from thoth.adviser.sieve import Sieve
from thoth.adviser.step import Step
from thoth.adviser.stride import Stride
from thoth.adviser.unit import Unit
from thoth.adviser.wrap import Wrap
from thoth.adviser.exceptions import PipelineConfigurationError
from thoth.adviser.exceptions import UnknownPipelineUnitError
from thoth.python import Project

import tests.units as units

from .base import AdviserTestCase
from .helpers import use_test_units


@pytest.fixture
def builder_context() -> PipelineBuilderContext:
    """Fixture for a builder context."""
    builder_context = PipelineBuilderContext(
        graph=None,
        project=None,
        library_usage=None,
        labels={},
        decision_type=DecisionType.RANDOM,
        recommendation_type=None,
    )
    builder_context.add_unit(units.boots.Boot1())
    builder_context.add_unit(units.pseudonyms.Pseudonym1())
    builder_context.add_unit(units.sieves.Sieve1())
    builder_context.add_unit(units.steps.Step1())
    builder_context.add_unit(units.strides.Stride1())
    builder_context.add_unit(units.wraps.Wrap1())
    return builder_context


class TestPipelineBuilderContext(AdviserTestCase):
    """Test context carried within pipeline builder."""

    def test_is_included(self, builder_context: PipelineBuilderContext) -> None:
        """Test check if the given pipeline unit is included."""
        assert builder_context.is_included(units.boots.Boot1)
        assert not builder_context.is_included(units.boots.Boot2)
        assert builder_context.is_included(units.pseudonyms.Pseudonym1)
        assert not builder_context.is_included(units.pseudonyms.Pseudonym2)
        assert builder_context.is_included(units.sieves.Sieve1)
        assert not builder_context.is_included(units.sieves.Sieve2)
        assert builder_context.is_included(units.steps.Step1)
        assert not builder_context.is_included(units.steps.Step2)
        assert builder_context.is_included(units.strides.Stride1)
        assert not builder_context.is_included(units.strides.Stride2)
        assert builder_context.is_included(units.wraps.Wrap1)
        assert not builder_context.is_included(units.wraps.Wrap2)

    def test_is_adviser_pipeline(self) -> None:
        """Test check for an adviser build context."""
        builder_context = PipelineBuilderContext(
            graph=None,
            project=None,
            library_usage=None,
            decision_type=None,
            recommendation_type=RecommendationType.LATEST,
        )
        assert builder_context.is_adviser_pipeline()
        assert not builder_context.is_dependency_monkey_pipeline()

    def test_is_dependency_monkey_pipeline(self) -> None:
        """Test check for a dependency monkey build context."""
        builder_context = PipelineBuilderContext(
            graph=None,
            project=None,
            library_usage=None,
            decision_type=DecisionType.RANDOM,
            recommendation_type=None,
        )
        assert builder_context.is_dependency_monkey_pipeline()
        assert not builder_context.is_adviser_pipeline()

    def test_invalid_pipeline_builder_context(self) -> None:
        """Test check an invalid build context."""
        # Exactly one from decision type/recommendation type has to be specified.
        with pytest.raises(ValueError):
            PipelineBuilderContext(
                graph=None,
                project=None,
                library_usage=None,
                decision_type=None,
                recommendation_type=None,
            )

        with pytest.raises(ValueError):
            PipelineBuilderContext(
                graph=None,
                project=None,
                library_usage=None,
                decision_type=DecisionType.ALL,
                recommendation_type=RecommendationType.LATEST,
            )

    @pytest.mark.parametrize(
        "unit_class,builder_context_attr",
        [
            (units.boots.Boot2, "boots"),
            (units.pseudonyms.Pseudonym2, "pseudonyms"),
            (units.sieves.Sieve2, "sieves"),
            (units.steps.Step2, "steps"),
            (units.strides.Stride2, "strides"),
            (units.wraps.Wrap2, "wraps"),
        ],
    )
    def test_add_unit(
        self,
        builder_context: PipelineBuilderContext,
        unit_class: Unit,
        builder_context_attr: str,
    ) -> None:
        """Test addition of a unit."""
        assert not builder_context.is_included(unit_class)
        unit = unit_class()
        builder_context.add_unit(unit)
        assert builder_context.is_included(unit_class)
        assert getattr(builder_context, builder_context_attr)[-1] is unit

    def test_get_included_boots(self) -> None:
        """Test get included boots of the provided boot class."""
        builder_context = PipelineBuilderContext(decision_type=DecisionType.RANDOM)
        assert list(builder_context.get_included_boots(units.boots.Boot1)) == []
        assert list(builder_context.get_included_boot_names()) == []
        unit = units.boots.Boot1()
        builder_context.add_unit(unit)
        assert list(builder_context.get_included_boots(units.boots.Boot1)) == [unit]
        assert list(builder_context.get_included_boot_names()) == ["Boot1"]

    def test_get_included_pseudonyms(self) -> None:
        """Test get included pseudonyms of the provided pseudonym class."""
        builder_context = PipelineBuilderContext(decision_type=DecisionType.RANDOM)
        assert list(builder_context.get_included_pseudonyms(units.pseudonyms.Pseudonym1)) == []
        assert list(builder_context.get_included_pseudonym_names()) == []
        unit = units.pseudonyms.Pseudonym1()
        builder_context.add_unit(unit)
        assert list(builder_context.get_included_pseudonyms(units.pseudonyms.Pseudonym1)) == [unit]
        assert list(builder_context.get_included_pseudonym_names()) == ["Pseudonym1"]

    def test_get_included_sieves(self) -> None:
        """Get included sieves of the provided sieve class."""
        builder_context = PipelineBuilderContext(decision_type=DecisionType.RANDOM)
        assert list(builder_context.get_included_sieves(units.sieves.Sieve1)) == []
        assert list(builder_context.get_included_sieve_names()) == []
        unit = units.sieves.Sieve1()
        builder_context.add_unit(unit)
        assert list(builder_context.get_included_sieves(units.sieves.Sieve1)) == [unit]
        assert list(builder_context.get_included_sieve_names()) == ["Sieve1"]

    def test_get_included_steps(self) -> None:
        """Get included steps of the provided step class."""
        builder_context = PipelineBuilderContext(decision_type=DecisionType.RANDOM)
        assert list(builder_context.get_included_steps(units.steps.Step1)) == []
        assert list(builder_context.get_included_step_names()) == []
        unit = units.steps.Step1()
        builder_context.add_unit(unit)
        assert list(builder_context.get_included_steps(units.steps.Step1)) == [unit]
        assert list(builder_context.get_included_step_names()) == ["Step1"]

    def test_get_included_strides(self) -> None:
        """Get included strides of the provided stride class."""
        builder_context = PipelineBuilderContext(decision_type=DecisionType.RANDOM)
        assert list(builder_context.get_included_strides(units.strides.Stride1)) == []
        assert list(builder_context.get_included_stride_names()) == []
        unit = units.strides.Stride1()
        builder_context.add_unit(unit)
        assert list(builder_context.get_included_strides(units.strides.Stride1)) == [unit]
        assert list(builder_context.get_included_stride_names()) == ["Stride1"]

    def test_get_included_wraps(self) -> None:
        """Get included wraps of the provided wrap class."""
        builder_context = PipelineBuilderContext(decision_type=DecisionType.RANDOM)
        assert list(builder_context.get_included_wraps(units.wraps.Wrap1)) == []
        assert list(builder_context.get_included_wrap_names()) == []
        unit = units.wraps.Wrap1()
        builder_context.add_unit(unit)
        assert list(builder_context.get_included_wraps(units.wraps.Wrap1)) == [unit]
        assert list(builder_context.get_included_wrap_names()) == ["Wrap1"]


class TestPipelineBuilder(AdviserTestCase):
    """Tests related to pipeline builder."""

    @use_test_units
    @pytest.mark.parametrize(
        "pipeline_config_method,kwargs",
        [
            (
                "get_adviser_pipeline_config",
                {"recommendation_type": RecommendationType.LATEST},
            ),
            (
                "get_dependency_monkey_pipeline_config",
                {"decision_type": DecisionType.RANDOM},
            ),
        ],
    )
    def test_build_configuration(
        self,
        pipeline_config_method: str,
        kwargs: Dict[str, Union[RecommendationType, DecisionType]],
    ) -> None:
        """Test building configuration."""
        # All test units do not register themselves - let's cherry-pick ones that should be present.
        # There are done 3 iterations in total during pipeline configuration creation.
        flexmock(units.boots.Boot1).should_receive("should_include").and_yield(
            {"some_parameter": 1.0}
        ).and_yield().and_yield().times(3)
        flexmock(units.pseudonyms.Pseudonym2).should_receive("should_include").and_yield(
            {}
        ).and_yield().and_yield().times(3)
        flexmock(units.sieves.Sieve2).should_receive("should_include").and_yield(
            {"foo": "bar"}
        ).and_yield().and_yield().times(3)
        flexmock(units.steps.Step1).should_receive("should_include").and_yield({}).and_yield().and_yield().times(3)
        flexmock(units.strides.Stride2).should_receive("should_include").and_yield({}).and_yield().and_yield().times(3)
        flexmock(units.strides.Stride1).should_receive("should_include").and_yield().and_yield(
            {"linus": "torvalds"}
        ).and_yield().times(3)
        flexmock(units.wraps.Wrap2).should_receive("should_include").and_yield({}).and_yield().and_yield().times(3)

        # It is not relevant if adviser/dependency monkey is called in this case.
        pipeline = getattr(PipelineBuilder, pipeline_config_method)(
            graph=None,
            project=None,
            library_usage=None,
            labels={},
            prescription=None,
            cli_parameters=None,
            **kwargs,
        )

        assert pipeline.to_dict() == {
            "boots": [
                {"name": "Boot1", "configuration": {"some_parameter": 1.0, "package_name": "flask"}, "unit_run": False}
            ],
            "sieves": [
                {
                    "name": "Sieve2",
                    "configuration": {"date": "2015-09-15", "foo": "bar", "package_name": "selinon"},
                    "unit_run": False,
                }
            ],
            "pseudonyms": [{"name": "Pseudonym2", "configuration": {"package_name": "flask"}, "unit_run": False}],
            "steps": [
                {
                    "name": "Step1",
                    "configuration": {
                        "guido_retirement": 2019,
                        "package_name": "tensorflow",
                        "multi_package_resolution": False,
                    },
                    "unit_run": False,
                }
            ],
            "strides": [
                {"name": "Stride2", "configuration": {"foo": None, "package_name": "thamos"}, "unit_run": False},
                {
                    "name": "Stride1",
                    "configuration": {
                        "linus": "torvalds",
                        "package_name": None,
                    },
                    "unit_run": False,
                },
            ],
            "wraps": [{"name": "Wrap2", "configuration": {"package_name": "hexsticker"}, "unit_run": False}],
        }

    @use_test_units
    def test_blocked_units(self, project: Project) -> None:
        """Test preventing a pipeline unit from being added to pipeline configuration."""
        flexmock(units.boots.Boot1).should_receive("should_include").and_yield({}).and_yield().times(2)
        flexmock(units.steps.Step1).should_receive("should_include").and_yield({}).and_yield().times(2)

        pipeline = PipelineBuilder.get_adviser_pipeline_config(
            recommendation_type=RecommendationType.LATEST,
            graph=None,
            project=project,
            labels={},
            library_usage=None,
            prescription=None,
            cli_parameters=None,
        )

        assert len(pipeline.boots) == 1
        assert pipeline.boots[0].__class__ is units.boots.Boot1

        assert len(pipeline.steps) == 1
        assert pipeline.steps[0].__class__ is units.steps.Step1

        assert "THOTH_ADVISER_BLOCKED_UNITS" not in os.environ

        try:
            os.environ["THOTH_ADVISER_BLOCKED_UNITS"] = f"{units.boots.Boot1.__name__},{units.steps.Step1.__name__}"

            flexmock(units.boots.Boot1).should_receive("should_include").times(0)
            flexmock(units.steps.Step1).should_receive("should_include").times(0)

            pipeline = PipelineBuilder.get_adviser_pipeline_config(
                recommendation_type=RecommendationType.LATEST,
                graph=None,
                project=None,
                labels={},
                library_usage=None,
                prescription=None,
                cli_parameters=None,
            )
            assert len(pipeline.boots) == 0
            assert len(pipeline.steps) == 0
        finally:
            os.environ.pop("THOTH_ADVISER_BLOCKED_UNITS")

    @use_test_units
    def test_from_dict(self) -> None:
        """Test instantiation of a pipeline from a dictionary."""
        dict_ = {
            "boots": [
                {"name": "Boot1", "configuration": {"some_parameter": 1.0, "package_name": "flask"}, "unit_run": False}
            ],
            "pseudonyms": [{"name": "Pseudonym2", "configuration": {"package_name": "tensorflow"}, "unit_run": False}],
            "sieves": [
                {
                    "name": "Sieve2",
                    "configuration": {"date": "2015-09-15", "foo": "bar", "package_name": "selinon"},
                    "unit_run": False,
                }
            ],
            "steps": [
                {
                    "name": "Step1",
                    "configuration": {
                        "guido_retirement": 2019,
                        "package_name": "tensorflow",
                        "multi_package_resolution": False,
                    },
                    "unit_run": False,
                }
            ],
            "strides": [
                {"name": "Stride2", "configuration": {"foo": None, "package_name": "thamos"}, "unit_run": False},
                {"name": "Stride1", "configuration": {"linus": "torvalds", "package_name": None}, "unit_run": False},
            ],
            "wraps": [{"name": "Wrap2", "configuration": {"package_name": "hexsticker"}, "unit_run": False}],
        }

        pipeline = PipelineBuilder.from_dict(dict(dict_))
        pipeline_dict = pipeline.to_dict()
        assert pipeline_dict == dict_
        assert isinstance(pipeline.boots[0], Boot)
        assert isinstance(pipeline.pseudonyms[0], Pseudonym)
        assert isinstance(pipeline.sieves[0], Sieve)
        assert isinstance(pipeline.steps[0], Step)
        assert isinstance(pipeline.strides[0], Stride)
        assert isinstance(pipeline.strides[1], Stride)
        assert isinstance(pipeline.wraps[0], Wrap)

    @use_test_units
    def test_load(self, tmp_path: Path) -> None:
        """Test instantiation of a pipeline from a dictionary."""
        expected_dict_ = {
            "boots": [
                {"configuration": {"package_name": "flask", "some_parameter": -0.2}, "name": "Boot1", "unit_run": False}
            ],
            "pseudonyms": [],
            "sieves": [],
            "steps": [
                {
                    "configuration": {
                        "guido_retirement": 2019,
                        "package_name": "tensorflow",
                        "multi_package_resolution": False,
                    },
                    "name": "Step1",
                    "unit_run": False,
                }
            ],
            "strides": [
                {"configuration": {"foo": None, "package_name": "thamos"}, "name": "Stride2", "unit_run": False}
            ],
            "wraps": [],
        }

        dict_ = {
            "boots": [{"name": "Boot1"}],
            "pseudonyms": [],
            "sieves": [],
            "steps": [
                {
                    "name": "Step1",
                    "unit_run": False,
                    "configuration": {"guido_retirement": 2019, "package_name": "tensorflow"},
                }
            ],
            "strides": [{"name": "Stride2", "unit_run": False}],
            "wraps": [],
        }

        yaml_path = tmp_path / "config.yaml"
        with open(str(yaml_path), "w") as f:
            yaml.safe_dump(dict_, f)

        pipeline = PipelineBuilder.load(yaml_path)

        assert isinstance(pipeline.boots[0], Boot)
        assert not pipeline.pseudonyms
        assert not pipeline.sieves
        assert isinstance(pipeline.steps[0], Step)
        assert isinstance(pipeline.strides[0], Stride)
        assert not pipeline.wraps
        pipeline_dict = pipeline.to_dict()
        assert pipeline_dict == expected_dict_

        json_path = tmp_path / "config.json"
        with open(str(json_path), "w") as f:
            json.dump(dict_, f)

        pipeline = PipelineBuilder.load(json_path)
        assert isinstance(pipeline.boots[0], Boot)
        assert not pipeline.pseudonyms
        assert not pipeline.sieves
        assert isinstance(pipeline.steps[0], Step)
        assert isinstance(pipeline.strides[0], Stride)
        assert not pipeline.wraps
        pipeline_dict = pipeline.to_dict()
        assert pipeline_dict == expected_dict_

    @use_test_units
    def test_from_dict_unit_configuration_error(self) -> None:
        """Test instantiation of a pipeline unit in case configuration errors.."""
        dict_ = {
            "boots": [{"name": "Boot1", "configuration": {"foo": "bar"}}],
        }

        flexmock(units.boots.Boot1)
        units.boots.Boot1.should_receive("update_configuration").with_args({"foo": "bar"}).and_raise(ValueError)

        with pytest.raises(PipelineConfigurationError):
            PipelineBuilder.from_dict(dict_)

    @use_test_units
    def test_from_dict_unit_not_exist_error(self) -> None:
        """Test instantiation of a pipeline from a dictionary in case of missing unit."""
        dict_ = {
            "sieves": [{"name": "SieveThatDoesNotExist", "configuration": {"foo": "bar"}}],
        }

        with pytest.raises(UnknownPipelineUnitError):
            PipelineBuilder.from_dict(dict_)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test state representation."""

import gc
import pytest
import random
import string
from itertools import chain

from thoth.adviser.state import State
from .base import AdviserTestCase


@pytest.fixture
def final_state() -> State:  # noqa: D401
    """A fixture for a final state."""
    state = State(
        score=0.5,
        resolved_dependencies={"daiquiri": ("daiquiri", "1.6.0", "https://pypi.org/simple")},
        unresolved_dependencies={},
        advised_runtime_environment=None,
    )
    state.add_justification(AdviserTestCase.JUSTIFICATION_SAMPLE_1)
    return state


class TestState(AdviserTestCase):
    """Test state representation and manipulation."""

    def test_justification(self, final_state: State) -> None:
        """Test manipulation with state justification."""
        assert final_state.justification == self.JUSTIFICATION_SAMPLE_1
        final_state.add_justification(self.JUSTIFICATION_SAMPLE_2)
        assert final_state.justification == list(chain(self.JUSTIFICATION_SAMPLE_1, self.JUSTIFICATION_SAMPLE_2))

    def test_is_final(self, state: State, final_state: State) -> None:
        """Test checks for final states."""
        assert final_state.is_final()
        assert not state.is_final()
        state.unresolved_dependencies.pop("flask")
        assert state.is_final()

    def test_add_unresolved_dependency(self, final_state: State) -> None:
        """Test adding unresolved dependencies into a state."""
        # This is actually never done in the code (from final state to a non-final state), but
        # let's test turning the switch.
        assert final_state.is_final()
        final_state.add_unresolved_dependency(("selinon", "1.0.0", "https://pypi.org/simple"))
        assert not final_state.is_final()
        final_state.unresolved_dependencies.pop("selinon")
        assert final_state.is_final()

    def test_get_random_first_unresolved_dependency(self) -> None:
        """Test getting random first unresolved dependency."""
        state = State(score=1.0)
        package_tuple1 = ("tensorflow", "2.1.0", "https://pypi.org/simple")
        package_tuple2 = ("selinon", "1.0.0", "https://pypi.org/simple")

        state.add_unresolved_dependency(package_tuple1)

        assert state.get_random_first_unresolved_dependency() is package_tuple1

        state.add_unresolved_dependency(package_tuple2)

        random_state = random.getstate()
        try:
            random.seed(42)
            assert state.get_random_first_unresolved_dependency() is package_tuple1
            assert state.get_random_first_unresolved_dependency() is package_tuple1
            assert state.get_random_first_unresolved_dependency() is package_tuple2
            assert state.get_random_first_unresolved_dependency() is package_tuple1
        finally:
            random.setstate(random_state)

    def test_get_random_first_unresolved_dependency_error(self) -> None:
        """Test raising an error on random first unresolved dependency if no dependency is available."""
        state = State(score=1.0)
        with pytest.raises(IndexError):
            state.get_random_first_unresolved_dependency()

    def test_clone(self, state: State) -> None:
        """Test cloning of states and their memory footprints."""
        cloned_state = state.clone()
        assert cloned_state is not state
        assert cloned_state.score == state.score

        # Check swallow copies.
        assert cloned_state.unresolved_dependencies is not state.unresolved_dependencies
        assert cloned_state.unresolved_dependencies == state.unresolved_dependencies

        for dependency_name in cloned_state.unresolved_dependencies:
            assert (
                cloned_state.unresolved_dependencies[dependency_name] == state.unresolved_dependencies[dependency_name]
            )
            assert (
                cloned_state.unresolved_dependencies[dependency_name]
                is not state.unresolved_dependencies[dependency_name]
            )

        assert cloned_state.resolved_dependencies is not state.resolved_dependencies
        assert cloned_state.resolved_dependencies == state.resolved_dependencies
        assert cloned_state.advised_runtime_environment is not state.advised_runtime_environment
        assert cloned_state.advised_runtime_environment == state.advised_runtime_environment
        assert cloned_state.justification is not state.justification
        assert cloned_state.justification == state.justification
        assert cloned_state.advised_manifest_changes is not state.advised_manifest_changes

    def test_parent(self) -> None:
        """Test referencing parent and weak reference handling."""
        state = State()
        cloned_state = state.clone()

        assert cloned_state.parent is state

        del state
        gc.collect()

        assert cloned_state.parent is None

    @pytest.mark.parametrize("n", range(5))
    def test_dict_order(self, n: int) -> None:
        """Test relative insertion into dict preserves order.

        This is an implementation detail for Python3.6 dict implementation and
        requirement for Python 3.7. Make sure the running interpreter obeys
        this.
        """
        items = []
        seen = set()
        for i in range(10):
            key = "".join(random.choice(string.ascii_letters) for i in range(n))
            if key in seen:
                continue

            seen.add(key)
            items.append((key, i))

        dict_ = dict(items)
        assert list(dict_.items()) == items, "Order of items in dict is not preserved"

        dict_["foo"] = 3003
        assert list(dict_.items())[-1] == ("foo", 3003), "Last item added to dict is not last obtained"



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of boot prescription v1."""

from typing import Optional

from flexmock import flexmock
import pytest
import yaml

from thoth.adviser.context import Context
from thoth.adviser.prescription.v1 import BootPrescription
from thoth.adviser.prescription.v1.schema import PRESCRIPTION_BOOT_SCHEMA

from .base import AdviserUnitPrescriptionTestCase


class TestBootPrescription(AdviserUnitPrescriptionTestCase):
    """Tests related to boot prescription v1."""

    def test_run_stack_info(self, context: Context) -> None:
        """Check assigning stack info."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_name: flask
run:
  stack_info:
    - type: ERROR
      message: "Unable to perform this operation"
      link: https://thoth-station.ninja
    - type: INFO
      message: "Hello, Thoth!"
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)
        self.check_run_stack_info(context, BootPrescription)

    def test_run_eager_stop_pipeline(self, context: Context) -> None:
        """Check eager stop pipeline configuration."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  dependency_monkey_pipeline: true
run:
  eager_stop_pipeline: This is exception message reported
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)
        self.check_run_eager_stop_pipeline(context, BootPrescription)

    def test_run_not_acceptable(self, context: Context) -> None:
        """Check raising not acceptable."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
  dependency_monkey_pipeline: true
run:
  not_acceptable: This is exception message reported
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)
        self.check_run_not_acceptable(context, BootPrescription)

    @pytest.mark.parametrize("log_level", ["INFO", "ERROR", "WARNING"])
    def test_run_log(self, caplog, context: Context, log_level: str) -> None:
        """Check reporting to log."""
        prescription_str = f"""
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
  dependency_monkey_pipeline: true
run:
  log:
    message: Some message logged
    type: {log_level}
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)
        self.check_run_log(caplog, context, log_level, BootPrescription)

    @pytest.mark.parametrize("package_name", [None, "flask"])
    def test_run_match(self, package_name: Optional[str]) -> None:
        """Test proper initialization based on yielded package_name."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_name: flask
run:
  stack_info:
    - type: ERROR
      message: "Unable to perform this operation"
      link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)
        unit = BootPrescription()
        unit.update_configuration({"package_name": package_name})
        assert unit.configuration["package_name"] == package_name

    def test_should_include_package_name(self) -> None:
        """Test including this pipeline unit."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
match:
  package_name: flask
run:
  stack_info:
    - type: ERROR
      message: "Unable to perform this operation"
      link: https://thoth-station.ninja
"""
        flexmock(BootPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(BootPrescription.should_include(builder_context)) == [
            {
                "package_name": "flask",
                "match": {
                    "package_name": "flask",
                },
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {
                            "type": "ERROR",
                            "message": "Unable to perform this operation",
                            "link": "https://thoth-station.ninja",
                        }
                    ]
                },
            }
        ]

    def test_should_include_multi(self) -> None:
        """Test including this pipeline unit multiple times."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
match:
  - package_name: flask
  - package_name: pandas
run:
  stack_info:
    - type: ERROR
      message: "Unable to perform this operation"
      link: https://thoth-station.ninja
"""
        flexmock(BootPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)

        builder_context = flexmock()
        result = list(BootPrescription.should_include(builder_context))
        assert len(result) == 2
        assert result == [
            {
                "package_name": "flask",
                "match": {
                    "package_name": "flask",
                },
                "prescription": {
                    "run": False,
                },
                "run": {
                    "stack_info": [
                        {
                            "type": "ERROR",
                            "message": "Unable to perform this operation",
                            "link": "https://thoth-station.ninja",
                        }
                    ]
                },
            },
            {
                "package_name": "pandas",
                "match": {
                    "package_name": "pandas",
                },
                "prescription": {
                    "run": False,
                },
                "run": {
                    "stack_info": [
                        {
                            "type": "ERROR",
                            "message": "Unable to perform this operation",
                            "link": "https://thoth-station.ninja",
                        }
                    ]
                },
            },
        ]
        assert result[0]["prescription"] is result[1]["prescription"]

    def test_should_include_no_package_name(self) -> None:
        """Test including this pipeline unit without package name."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
run:
  stack_info:
    - type: INFO
      message: Yet another text printed.
      link: https://thoth-station.ninja
"""
        flexmock(BootPrescription).should_receive("_should_include_base").replace_with(lambda _: True).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(BootPrescription.should_include(builder_context)) == [
            {
                "match": {},
                "package_name": None,
                "prescription": {"run": False},
                "run": {
                    "stack_info": [
                        {
                            "type": "INFO",
                            "message": "Yet another text printed.",
                            "link": "https://thoth-station.ninja",
                        }
                    ]
                },
            }
        ]

    def test_no_should_include(self) -> None:
        """Test including this pipeline unit without package name."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
run:
  stack_info:
    - type: INFO
      message: Yet another text printed.
      link: https://thoth-station.ninja
"""
        flexmock(BootPrescription).should_receive("_should_include_base").replace_with(lambda _: False).once()
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)

        builder_context = flexmock()
        assert list(BootPrescription.should_include(builder_context)) == []

    def test_should_include_multi_run(self, context: Context) -> None:
        """Test including this pipeline unit multiple times."""
        prescription_str = """
name: BootUnit
type: boot
should_include:
  times: 1
  adviser_pipeline: true
match:
- package_name: flask
- package_name: pandas
run:
  stack_info:
  - type: INFO
    message: "Hello, Thoth!"
    link: https://thoth-station.ninja
"""
        prescription = yaml.safe_load(prescription_str)
        PRESCRIPTION_BOOT_SCHEMA(prescription)
        BootPrescription.set_prescription(prescription)
        unit = BootPrescription()

        assert "flask" in (pv.name for pv in context.project.iter_dependencies())

        unit.pre_run()
        with unit.assigned_context(context):
            # Run twice.
            unit.run()
            unit.run()

        assert context.stack_info == [
            {"type": "INFO", "message": "Hello, Thoth!", "link": "https://thoth-station.ninja"}
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing strides."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr
from voluptuous import Any as SchemaAny
from voluptuous import Schema
from voluptuous import Required

from thoth.adviser.state import State
from .unit import UnitPrescription
from .schema import PRESCRIPTION_STRIDE_RUN_SCHEMA
from .schema import PRESCRIPTION_STRIDE_MATCH_ENTRY_SCHEMA

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext


@attr.s(slots=True)
class StridePrescription(UnitPrescription):
    """Stride base class implementation."""

    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): SchemaAny(str, None),
            Required("match"): PRESCRIPTION_STRIDE_MATCH_ENTRY_SCHEMA,
            Required("run"): PRESCRIPTION_STRIDE_RUN_SCHEMA,
            Required("prescription"): Schema({"run": bool}),
        }
    )

    @staticmethod
    def is_stride_unit_type() -> bool:
        """Check if this unit is of type stride."""
        return True

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if the given pipeline unit should be included in the given pipeline configuration."""
        if cls._should_include_base(builder_context):
            prescription: Dict[str, Any] = cls._PRESCRIPTION  # type: ignore
            yield from cls._yield_should_include_with_state(prescription)
            return None

        yield from ()
        return None

    def run(self, state: State) -> None:
        """Run main entry-point for strides."""
        if not self._run_state(state):
            return None

        try:
            self._run_base()
        finally:
            self._configuration["prescription"]["run"] = True



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Definition of package resolution based on precomputed data available in graph.

There are defined primitives required for offline resolution. This off-line
resolution is done based on data aggregated in the graph database so thereis
no need to perform resolving by actually installing Python packages (as this
version resolution is dynamic in case of Python).
"""

from typing import List
from typing import Dict
from typing import Generator
from typing import Tuple
from typing import Set

import attr
from packaging.requirements import Requirement
from thoth.common import RuntimeEnvironment
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase
from thoth.solver.python.base import DependencyParser
from thoth.solver.python.base import ReleasesFetcher
from thoth.solver.python.base import Solver
from thoth.solver.python import PythonDependencyParser


@attr.s(slots=True)
class GraphReleasesFetcher(ReleasesFetcher):  # type: ignore
    """Fetch releases for packages from the graph database."""

    graph = attr.ib(type=GraphDatabase, kw_only=True)
    runtime_environment = attr.ib(
        type=RuntimeEnvironment,
        default=attr.Factory(RuntimeEnvironment.from_dict),
        kw_only=True,
    )

    def fetch_releases(self, package_name: str) -> Tuple[str, List[Tuple[str, str]]]:
        """Fetch releases for the given package name."""
        # Make sure we have normalized names in the graph database according to PEP:
        #   https://www.python.org/dev/peps/pep-0503/#normalized-names
        package_name = Source.normalize_package_name(package_name)

        start_offset = 0
        result: Set[Tuple[str, str, str]] = set()
        while True:
            query_result = self.graph.get_solved_python_package_versions_all(
                package_name=package_name,
                os_name=self.runtime_environment.operating_system.name,
                os_version=self.runtime_environment.operating_system.version,
                python_version=self.runtime_environment.python_version,
                start_offset=start_offset,
                count=self.graph.DEFAULT_COUNT,
                distinct=True,
                is_missing=False,
            )

            start_offset += 1
            result.update(query_result)

            # We have reached end of pagination or no versions were found.
            if len(query_result) < self.graph.DEFAULT_COUNT:
                break

        return package_name, [(version, index_url) for _, version, index_url in result]


@attr.s(slots=True)
class PackageVersionDependencyParser(DependencyParser):  # type: ignore
    """Parse an instance of PackageVersion to Dependency object needed by solver."""

    def parse(self, dependencies: List[PackageVersion]) -> Generator[Requirement, None, None]:
        """Parse the given list of PackageVersion objects."""
        for package_version in dependencies:
            version = package_version.version if package_version.version not in ("*", None) else ""
            dependency = PythonDependencyParser.parse_python(package_version.name + version)
            yield dependency


@attr.s(slots=True)
class PythonGraphSolver(Solver):  # type: ignore
    """Solve Python dependencies based on data available in the graph database."""

    dependency_parser = attr.ib(type=PackageVersionDependencyParser, kw_only=True)
    releases_fetcher = attr.ib(type=GraphReleasesFetcher, kw_only=True)


@attr.s(slots=True)
class PythonPackageGraphSolver:
    """A wrapper to manipulate with Python packages using pure PackageVersion object interface."""

    graph = attr.ib(type=GraphDatabase, kw_only=True)
    runtime_environment = attr.ib(
        type=RuntimeEnvironment,
        kw_only=True,
        default=attr.Factory(RuntimeEnvironment.from_dict),
    )
    # Do not instantiate multiple objects for same python package tuple to optimize memory usage.
    _package_versions = attr.ib(
        type=Dict[Tuple[str, str, str], PackageVersion],
        default=attr.Factory(dict),
        kw_only=True,
    )
    # Have just one instance of Source object per python package source index url.
    _sources = attr.ib(type=Dict[str, Source], default=attr.Factory(dict), kw_only=True)
    _solver = attr.ib(type=PythonGraphSolver, default=None, kw_only=True)

    @property
    def solver(self) -> PythonGraphSolver:
        """Retrieve solver instance resolving using graph database."""
        if not self._solver:
            self._solver = PythonGraphSolver(
                dependency_parser=PackageVersionDependencyParser(),
                releases_fetcher=GraphReleasesFetcher(graph=self.graph, runtime_environment=self.runtime_environment),
            )

        return self._solver

    def solve(self, dependencies: List[PackageVersion], graceful: bool = True) -> Dict[str, List[PackageVersion]]:
        """Solve the given dependencies and return object representation of packages."""
        result = {}
        # First, construct the map for checking packages.
        dependencies_map = {dependency.name: dependency for dependency in dependencies}

        resolved = self.solver.solve(dependencies, graceful=graceful)
        if not resolved:
            return {}

        for package_name, versions in resolved.items():
            # If this pop fails, it means that the package name has changed over the resolution.
            original_package = dependencies_map.pop(package_name)
            result_versions = []
            for version, index_url in versions:
                package_tuple = (original_package.name, version, index_url)
                package_version = self._package_versions.get(package_tuple)
                if not package_version:
                    source = self._sources.get(index_url)
                    if not source:
                        source = Source(index_url)
                        self._sources[index_url] = source

                    package_version = PackageVersion(
                        name=original_package.name,
                        version="==" + version,
                        index=source,
                        develop=original_package.develop,
                    )

                result_versions.append(package_version)

            result[original_package.name] = result_versions

        return result



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test filtering out packages based on enabled or disabled Python package index."""

from flexmock import flexmock

from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import PackageIndexSieve
from thoth.python import PackageVersion
from thoth.python import Source
from thoth.storages import GraphDatabase
from thoth.storages.exceptions import NotFoundError

from ..base import AdviserUnitTestCase


class TestPackageIndexSieve(AdviserUnitTestCase):
    """Test filtering out packages based on enabled or disabled Python package index."""

    UNIT_TESTED = PackageIndexSieve

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_sieve_index_enabled(self) -> None:
        """Test no-op when the given Python package index used to obtain package is enabled."""
        source = Source("https://pypi.org/simple")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("is_python_package_index_enabled").with_args(source.url).and_return(True).once()

        context = flexmock(graph=GraphDatabase())
        with PackageIndexSieve.assigned_context(context):
            sieve = PackageIndexSieve()
            assert list(sieve.run((p for p in [package_version]))) == [package_version]

    def test_sieve_index_disabled(self) -> None:
        """Test removals of Python package if Python package index used is disabled."""
        source = Source("https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/jemalloc/simple/")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("is_python_package_index_enabled").with_args(source.url).and_return(False).once()

        context = flexmock(graph=GraphDatabase())
        with PackageIndexSieve.assigned_context(context):
            sieve = PackageIndexSieve()
            assert list(sieve.run(p for p in [package_version])) == []

    def test_sieve_index_not_found(self) -> None:
        """Test removals of Python package if Python package index used is unknown."""
        source = Source("https://tensorflow.pypi.thoth-station.ninja/index/manylinux2010/jemalloc/simple/")
        package_version = PackageVersion(name="tensorflow", version="==1.9.0", index=source, develop=False)
        flexmock(GraphDatabase)
        GraphDatabase.should_receive("is_python_package_index_enabled").with_args(source.url).and_raise(
            NotFoundError
        ).once()

        context = flexmock(graph=GraphDatabase())
        with PackageIndexSieve.assigned_context(context):
            sieve = PackageIndexSieve()
            assert list(sieve.run((p for p in [package_version]))) == []



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test adviser's context passed to pipeline units."""

from typing import Tuple

import pytest

from thoth.python import PackageVersion
from thoth.python import Source

from thoth.adviser.state import State
from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.enums import DecisionType
from thoth.adviser.exceptions import NotFound

from .base import AdviserTestCase


@pytest.fixture
def package_version() -> PackageVersion:  # noqa: D401
    """A fixture for a package version representative."""
    return PackageVersion(
        name="selinon",
        version="==1.0.0",
        index=Source("https://pypi.org/simple"),
        develop=False,
    )


@pytest.fixture
def package_tuple() -> Tuple[str, str, str]:  # noqa: D401
    """A fixture for a package tuple representative."""
    return "selinon", "1.0.0", "https://pypi.org/simple"


class TestContext(AdviserTestCase):
    """Test context carried within resolution."""

    def test_get_package_version(self, context: Context, package_version: PackageVersion) -> None:
        """Test getting registering and getting a package version."""
        with pytest.raises(NotFound):
            context.get_package_version(package_version.to_tuple())

        assert context.register_package_version(package_version) is False
        assert context.get_package_version(package_version.to_tuple()) is package_version

    def test_get_package_version_graceful(self, context: Context, package_version: PackageVersion) -> None:
        """Test getting registered package version, gracefully."""
        assert context.get_package_version(package_version.to_tuple(), graceful=True) is None
        with pytest.raises(NotFound):
            context.get_package_version(package_version.to_tuple(), graceful=False)

        assert context.register_package_version(package_version) is False

        assert context.get_package_version(package_version.to_tuple(), graceful=True) is package_version
        assert context.get_package_version(package_version.to_tuple(), graceful=False) is package_version

    def test_get_top_accepted_final_state(self, context: Context) -> None:
        """Test retrieval of top accepted final state."""
        context.count = 10

        assert context.get_top_accepted_final_state() is None

        state1 = State(score=0.0)
        assert context.register_accepted_final_state(state1) is None
        assert context.get_top_accepted_final_state() is state1

        state2 = State(score=1.0)
        assert context.register_accepted_final_state(state2) is None
        assert context.get_top_accepted_final_state() is state2

        state3 = State(score=2.0)
        assert context.register_accepted_final_state(state3) is None
        assert context.get_top_accepted_final_state() is state3

        state4 = State(score=0.5)
        assert context.register_accepted_final_state(state4) is None
        assert context.get_top_accepted_final_state() is state3

    def test_register_accepted_final_state(self, context: Context) -> None:
        """Test registering accepted final state and final state manipulation."""
        context.count = 2

        state1 = State(score=0.0)
        assert context.register_accepted_final_state(state1) is None
        assert state1 in context.iter_accepted_final_states()
        assert list(context.iter_accepted_final_states()) == [state1]

        state2 = State(score=1.0)
        assert context.register_accepted_final_state(state2) is None
        assert state1 in context.iter_accepted_final_states()
        assert state2 in context.iter_accepted_final_states()
        assert list(context.iter_accepted_final_states_sorted()) == [state2, state1]
        assert list(context.iter_accepted_final_states_sorted(reverse=True)) == [state2, state1]
        assert list(context.iter_accepted_final_states_sorted(reverse=False)) == [state1, state2]

        state3 = State(score=3.0)
        assert context.register_accepted_final_state(state3) is None
        assert state3 in context.iter_accepted_final_states_sorted()
        assert state2 in context.iter_accepted_final_states_sorted()
        assert list(context.iter_accepted_final_states_sorted()) == [state3, state2]
        assert list(context.iter_accepted_final_states_sorted(reverse=True)) == [state3, state2]
        assert list(context.iter_accepted_final_states_sorted(reverse=False)) == [state2, state3]

        state4 = State(score=2.0)
        assert context.register_accepted_final_state(state4) is None
        assert state3 in context.iter_accepted_final_states()
        assert state4 in context.iter_accepted_final_states()
        assert list(context.iter_accepted_final_states_sorted()) == [state3, state4]
        assert list(context.iter_accepted_final_states_sorted(reverse=True)) == [state3, state4]
        assert list(context.iter_accepted_final_states_sorted(reverse=False)) == [state4, state3]

        state5 = State(score=0.1)
        assert context.register_accepted_final_state(state5) is None
        assert state3 in context.iter_accepted_final_states()
        assert state4 in context.iter_accepted_final_states()
        assert list(context.iter_accepted_final_states_sorted()) == [state3, state4]
        assert list(context.iter_accepted_final_states_sorted(reverse=True)) == [state3, state4]
        assert list(context.iter_accepted_final_states_sorted(reverse=False)) == [state4, state3]

    def test_register_package_version_existing(self, context: Context, package_version: PackageVersion) -> None:
        """Test registering an existing package version to context."""
        assert context.register_package_version(package_version) is False
        assert context.get_package_version(package_version.to_tuple()) is package_version
        assert context.register_package_version(package_version) is True

    def test_register_package_tuple_new(self, context: Context, package_tuple: Tuple[str, str, str]) -> None:
        """Test registering a new package tuple to the context."""
        with pytest.raises(NotFound):
            context.get_package_version(package_tuple)

        extras = ["postgresql"]

        assert (
            context.register_package_tuple(
                package_tuple, develop=True, extras=extras, os_name="rhel", os_version="8.1", python_version="3.6"
            )
            is not None
        )

        package_version = context.get_package_version(package_tuple)

        assert package_version.name == "selinon"
        assert package_version.version == "==1.0.0"
        assert package_version.develop is True
        assert package_version.index is not None
        assert package_version.index.url == "https://pypi.org/simple"
        assert package_version.markers is None
        assert package_version.extras == extras

    def test_register_package_tuple_existing(self, context: Context, package_tuple: Tuple[str, str, str]) -> None:
        """Check registering an existing package tuple does not instantiate a new one."""
        with pytest.raises(NotFound):
            context.get_package_version(package_tuple)

        extras = ["postgresql"]

        package_version_registered = context.register_package_tuple(
            package_tuple, develop=True, extras=extras, os_name="fedora", os_version="31", python_version="3.7"
        )

        assert package_version_registered is not None

        package_version_another = context.register_package_tuple(
            package_tuple, develop=True, extras=extras, os_name="fedora", os_version="31", python_version="3.7"
        )

        assert package_version_registered is package_version_another, "Different instances returned"

    def test_note_dependencies(self, context: Context) -> None:
        """Test noting dependencies to the context."""
        dependency_tuple = ("tensorboard", "2.1.0", "https://pypi.org/simple")
        package_tuple = ("tensorflow", "2.0.0", "https://pypi.org/simple")

        context.register_package_version(
            PackageVersion(
                name=package_tuple[0],
                version="==" + package_tuple[1],
                index=Source(package_tuple[2]),
                develop=False,
            )
        )

        context.register_package_tuple(
            dependency_tuple,
            develop=True,
            extras=None,
            dependent_tuple=package_tuple,
            os_name="fedora",
            os_version="31",
            python_version="3.7",
        )

        package_version = context.get_package_version(package_tuple)
        assert package_version is not None
        assert package_version.name == package_tuple[0]
        assert package_version.locked_version == package_tuple[1]
        assert package_version.index.url == package_tuple[2]

        package_version = context.get_package_version(dependency_tuple)
        assert package_version is not None
        assert package_version.name == dependency_tuple[0]
        assert package_version.locked_version == dependency_tuple[1]
        assert package_version.index.url == dependency_tuple[2]

        assert package_tuple[0] in context.dependencies
        assert package_tuple in context.dependencies[package_tuple[0]]
        assert dependency_tuple in context.dependencies[package_tuple[0]][package_tuple]

        assert dependency_tuple[0] in context.dependents
        assert dependency_tuple in context.dependents[dependency_tuple[0]]
        entry = context.dependents[dependency_tuple[0]][dependency_tuple]
        assert entry == {(package_tuple, "fedora", "31", "3.7")}

        # By calling register_package_version we get a notion about direct dependency.
        assert package_tuple[0] in context.dependents
        assert package_tuple in context.dependents[package_tuple[0]]
        assert context.dependents[package_tuple[0]][package_tuple] == set()

    def test_is_dependency_monkey(self) -> None:
        """Test checking if the given context is an adviser context."""
        context = Context(
            project=None,
            graph=None,
            library_usage=None,
            labels={},
            limit=None,
            count=None,
            beam=None,
            recommendation_type=None,
            decision_type=DecisionType.ALL,
        )

        assert context.is_dependency_monkey() is True
        assert context.is_adviser() is False

    def test_is_adviser(self) -> None:
        """Test checking if the given context is an adviser context."""
        context = Context(
            project=None,
            graph=None,
            library_usage=None,
            labels={},
            limit=None,
            count=None,
            beam=None,
            recommendation_type=RecommendationType.LATEST,
            decision_type=None,
        )

        assert context.is_adviser() is True
        assert context.is_dependency_monkey() is False

    def test_invalid_context_no_type(self) -> None:
        """Exactly one type (recommendation/decision) has to be provided on instantiation."""
        with pytest.raises(ValueError):
            Context(
                project=None,
                graph=None,
                library_usage=None,
                labels={},
                limit=None,
                count=None,
                beam=None,
                # Decision type and recommendation type cannot be set to None at the same time.
                recommendation_type=None,
                decision_type=None,
            )

    def test_invalid_context_type(self) -> None:
        """Context type cannot capture both adviser and dependency monkey type."""
        with pytest.raises(ValueError):
            Context(
                project=None,
                graph=None,
                library_usage=None,
                limit=None,
                labels={},
                count=None,
                beam=None,
                # Decision type and recommendation type cannot be set at the same time.
                recommendation_type=RecommendationType.LATEST,
                decision_type=DecisionType.ALL,
            )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Schema v1 based prescription units."""

from .add_package_step import AddPackageStepPrescription
from .boot import BootPrescription
from .gh_release_notes import GHReleaseNotesWrapPrescription
from .group import GroupStepPrescription
from .pseudonym import PseudonymPrescription
from .sieve import SievePrescription
from .skip_package_sieve import SkipPackageSievePrescription
from .skip_package_step import SkipPackageStepPrescription
from .step import StepPrescription
from .stride import StridePrescription
from .unit import UnitPrescription
from .wrap import WrapPrescription


__all__ = [
    "AddPackageStepPrescription",
    "BootPrescription",
    "GHReleaseNotesWrapPrescription",
    "GroupStepPrescription",
    "PseudonymPrescription",
    "SievePrescription",
    "SkipPackageSievePrescription",
    "SkipPackageStepPrescription",
    "StepPrescription",
    "StridePrescription",
    "UnitPrescription",
    "WrapPrescription",
]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to provide information about the last CVE update."""

import logging
import os
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.common import datetime2datetime_str
from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class CveTimestampBoot(Boot):
    """A boot to provide information about the last CVE update."""

    _CVE_TIMESTAMP_DAYS_WARNING = int(os.getenv("THOTH_ADVISER_CVE_TIMESTAMP_DAYS_WARNING", 7))

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, once, always."""
        if builder_context.is_adviser_pipeline() and not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Check CVE timestamp and provide it to users."""
        cve_timestamp = self.context.graph.get_cve_timestamp()
        if cve_timestamp is None:
            message = "No CVE timestamp information found in the database"
            self.context.stack_info.append(
                {
                    "message": message,
                    "type": "WARNING",
                    "link": jl("no_cve_timestamp"),
                }
            )
            return

        days = (datetime.utcnow() - cve_timestamp).days
        self.context.stack_info.append(
            {
                "type": "WARNING" if days > self._CVE_TIMESTAMP_DAYS_WARNING else "INFO",
                "message": f"CVE database of known vulnerabilities for Python packages was "
                f"updated at {datetime2datetime_str(cve_timestamp)!r}",
                "link": jl("cve_timestamp"),
            }
        )
